"""
TransactionMetaPlugin offers a way of saving key-value data for transations.
You can use the plugin in same way as other plugins::


    meta_plugin = TransactionMetaPlugin()

    versioning_manager.plugins.add(meta_plugin)


TransactionMetaPlugin creates a simple model called TransactionMeta. This class
has three columns: transaction_id, key and value. TransactionMeta plugin also
creates an association proxy between TransactionMeta and Transaction classes
for easy dictionary based access of key-value pairs.

You can easily 'tag' transactions with certain key value pairs by giving these
keys and values to the meta property of Transaction class.


::


    from sqlalchemy_continuum import versioning_manager


    article = Article()
    session.add(article)

    uow = versioning_manager.unit_of_work(session)
    tx = uow.create_transaction(session)
    tx.meta = {u'some_key': u'some value'}
    session.commit()

    TransactionMeta = meta_plugin.model_class
    Transaction = versioning_manager.transaction_cls

    # find all transactions with 'article' tags
    query = (
        session.query(Transaction)
        .join(Transaction.meta_relation)
        .filter(
            db.and_(
                TransactionMeta.key == 'some_key',
                TransactionMeta.value == 'some value'
            )
        )
    )
"""

import sqlalchemy as sa
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.ext.associationproxy import association_proxy

from .base import Plugin
from ..factory import ModelFactory


class TransactionMetaBase(object):
    transaction_id = sa.Column(
        sa.BigInteger,
        primary_key=True
    )
    key = sa.Column(sa.Unicode(255), primary_key=True)
    value = sa.Column(sa.UnicodeText)


class TransactionMetaFactory(ModelFactory):
    model_name = 'TransactionMeta'

    def create_class(self, manager):
        """
        Create TransactionMeta class.
        """
        class TransactionMeta(
            manager.declarative_base,
            TransactionMetaBase
        ):
            __tablename__ = 'transaction_meta'

        TransactionMeta.transaction = sa.orm.relationship(
            manager.transaction_cls,
            backref=sa.orm.backref(
                'meta_relation',
                collection_class=attribute_mapped_collection('key')
            ),
            primaryjoin=(
                '%s.id == TransactionMeta.transaction_id' %
                manager.transaction_cls.__name__
            ),
            foreign_keys=[TransactionMeta.transaction_id]
        )

        manager.transaction_cls.meta = association_proxy(
            'meta_relation',
            'value',
            creator=lambda key, value: TransactionMeta(key=key, value=value)
        )

        return TransactionMeta


class TransactionMetaPlugin(Plugin):
    def after_build_tx_class(self, manager):
        self.model_class = TransactionMetaFactory()(manager)
        manager.transaction_meta_cls = self.model_class

    def after_build_models(self, manager):
        self.model_class = TransactionMetaFactory()(manager)
        manager.transaction_meta_cls = self.model_class



import sqlalchemy as sa

from .exc import ClassNotVersioned
from .expression_reflector import VersionExpressionReflector
from .operation import Operation
from .table_builder import TableBuilder
from .utils import adapt_columns, version_class, option


class RelationshipBuilder(object):
    def __init__(self, versioning_manager, model, property_):
        self.manager = versioning_manager
        self.property = property_
        self.model = model

    def one_to_many_subquery(self, obj):
        tx_column = option(obj, 'transaction_column_name')

        remote_alias = sa.orm.aliased(self.remote_cls)
        primary_keys = [
            getattr(remote_alias, column.name) for column
            in sa.inspect(remote_alias).mapper.columns
            if column.primary_key and column.name != tx_column
        ]

        return sa.exists(
            sa.select(
                [1]
            ).where(
                sa.and_(
                    getattr(remote_alias, tx_column) <=
                    getattr(obj, tx_column),
                    *[
                        getattr(remote_alias, pk.name) ==
                        getattr(self.remote_cls, pk.name)
                        for pk in primary_keys
                    ]
                )
            ).group_by(
                *primary_keys
            ).having(
                sa.func.max(getattr(remote_alias, tx_column)) ==
                getattr(self.remote_cls, tx_column)
            ).correlate(self.local_cls, self.remote_cls)
        )

    def many_to_one_subquery(self, obj):
        tx_column = option(obj, 'transaction_column_name')
        reflector = VersionExpressionReflector(obj, self.property)
        subquery = sa.select(
            [sa.func.max(getattr(self.remote_cls, tx_column))]
        ).where(
            sa.and_(
                getattr(self.remote_cls, tx_column) <=
                getattr(obj, tx_column),
                reflector(self.property.primaryjoin)
            )
        )
        try:
            subquery = subquery.scalar_subquery()
        except AttributeError:  # SQLAlchemy < 1.4
            subquery = subquery.as_scalar()

        return getattr(self.remote_cls, tx_column) == subquery

    def query(self, obj):
        session = sa.orm.object_session(obj)
        return (
            session.query(self.remote_cls)
            .filter(
                self.criteria(obj)
            )
        )

    def process_query(self, query):
        """
        Process given SQLAlchemy Query object depending on the associated
        RelationshipProperty object.

        :param query: SQLAlchemy Query object
        """
        if self.property.lazy == 'dynamic':
            return query
        if self.property.uselist is False:
            return query.first()
        return query.all()

    def criteria(self, obj):
        direction = self.property.direction

        if self.versioned:
            if direction.name == 'ONETOMANY':
                return self.one_to_many_criteria(obj)
            elif direction.name == 'MANYTOMANY':
                return self.many_to_many_criteria(obj)
            elif direction.name == 'MANYTOONE':
                return self.many_to_one_criteria(obj)
        else:
            reflector = VersionExpressionReflector(obj, self.property)
            return reflector(self.property.primaryjoin)

    def many_to_many_criteria(self, obj):
        """
        Returns the many-to-many query.

        Looks up remote items through associations and for each item returns
        returns the last version with a transaction less than or equal to the
        transaction of `obj`. This must hold true for both the association and
        the remote relation items.

        Example
        -------
        Select all tags of article with id 3 and transaction 5

        .. code-block:: sql

        SELECT tags_version.*
        FROM tags_version
        WHERE EXISTS (
            SELECT 1
            FROM article_tag_version
            WHERE article_id = 3
            AND tag_id = tags_version.id
            AND operation_type != 2
            AND EXISTS (
                SELECT 1
                FROM article_tag_version as article_tag_version2
                WHERE article_tag_version2.tag_id = article_tag_version.tag_id
                AND article_tag_version2.tx_id <= 5
                GROUP BY article_tag_version2.tag_id
                HAVING
                    MAX(article_tag_version2.tx_id) =
                    article_tag_version.tx_id
            )
        )
        AND EXISTS (
            SELECT 1
            FROM tags_version as tags_version_2
            WHERE tags_version_2.id = tags_version.id
            AND tags_version_2.tx_id <= 5
            GROUP BY tags_version_2.id
            HAVING MAX(tags_version_2.tx_id) = tags_version.tx_id
        )
        AND operation_type != 2
        """
        return sa.and_(
            self.association_subquery(obj),
            self.one_to_many_subquery(obj),
            self.remote_cls.operation_type != Operation.DELETE
        )

    def many_to_one_criteria(self, obj):
        """Returns the many-to-one query.

        Returns the item on the 'one' side with the highest transaction id
        as long as it is less or equal to the transaction id of the `obj`.

        Example
        -------
        Look up the Article of a Tag with article_id = 4 and
        transaction_id = 5

        .. code-block:: sql

        SELECT *
        FROM articles_version
        WHERE id = 4
        AND transaction_id = (
            SELECT max(transaction_id)
            FROM articles_version
            WHERE transaction_id <= 5
            AND id = 4
        )
        AND operation_type != 2

        """
        reflector = VersionExpressionReflector(obj, self.property)
        return sa.and_(
            reflector(self.property.primaryjoin),
            self.many_to_one_subquery(obj),
            self.remote_cls.operation_type != Operation.DELETE
        )

    def one_to_many_criteria(self, obj):
        """
        Returns the one-to-many query.

        For each item on the 'many' side, returns its latest version as long as
        the transaction of that version is less than equal of the transaction
        of `obj`.

        Example
        -------
        Using the Article-Tags relationship, where we look for tags of
        article_version with id = 3 and transaction = 5 the sql produced is

        .. code-block:: sql

        SELECT tags_version.*
        FROM tags_version
        WHERE tags_version.article_id = 3
        AND tags_version.operation_type != 2
        AND EXISTS (
            SELECT 1
            FROM tags_version as tags_version_last
            WHERE tags_version_last.transaction_id <= 5
            AND tags_version_last.id = tags_version.id
            GROUP BY tags_version_last.id
            HAVING
                MAX(tags_version_last.transaction_id) =
                tags_version.transaction_id
        )

        """
        reflector = VersionExpressionReflector(obj, self.property)
        return sa.and_(
            reflector(self.property.primaryjoin),
            self.one_to_many_subquery(obj),
            self.remote_cls.operation_type != Operation.DELETE
        )

    @property
    def reflected_relationship(self):
        """
        Builds a reflected one-to-many, one-to-one and many-to-one
        relationship between two version classes.
        """
        @property
        def relationship(obj):
            query = self.query(obj)
            return self.process_query(query)
        return relationship

    def association_subquery(self, obj):
        """
        Returns an EXISTS clause that checks if an association exists for given
        SQLAlchemy declarative object. This query is used by
        many_to_many_criteria method.

        Example query:

        .. code-block:: sql

        EXISTS (
            SELECT 1
            FROM article_tag_version
            WHERE article_id = 3
            AND tag_id = tags_version.id
            AND operation_type != 2
            AND EXISTS (
                SELECT 1
                FROM article_tag_version as article_tag_version2
                WHERE article_tag_version2.tag_id = article_tag_version.tag_id
                AND article_tag_version2.tx_id <=5
                AND article_tag_version2.article_id = 3
                GROUP BY article_tag_version2.tag_id
                HAVING
                    MAX(article_tag_version2.tx_id) =
                    article_tag_version.tx_id
            )
        )

        :param obj: SQLAlchemy declarative object
        """

        tx_column = option(obj, 'transaction_column_name')
        join_column = self.property.primaryjoin.right.name
        object_join_column = self.property.primaryjoin.left.name
        reflector = VersionExpressionReflector(obj, self.property)

        association_table_alias = self.association_version_table.alias()
        association_cols = [
            association_table_alias.c[association_col.name]
            for _, association_col
            in self.remote_to_association_column_pairs
        ]

        association_exists = sa.exists(
            sa.select(
                [1]
            ).where(
                sa.and_(
                    association_table_alias.c[tx_column] <=
                    getattr(obj, tx_column),
                    association_table_alias.c[join_column] == getattr(obj, object_join_column),
                    *[association_col ==
                      self.association_version_table.c[association_col.name]
                      for association_col
                      in association_cols]
                )
            ).group_by(
                *association_cols
            ).having(
                sa.func.max(association_table_alias.c[tx_column]) ==
                self.association_version_table.c[tx_column]
            ).correlate(self.association_version_table)
        )
        return sa.exists(
            sa.select(
                [1]
            ).where(
                sa.and_(
                    reflector(self.property.primaryjoin),
                    association_exists,
                    self.association_version_table.c.operation_type !=
                    Operation.DELETE,
                    adapt_columns(self.property.secondaryjoin),
                )
            ).correlate(self.local_cls, self.remote_cls)
        )

    def build_association_version_tables(self):
        """
        Builds many-to-many association version table for given property.
        Association version tables are used for tracking change history of
        many-to-many associations.
        """
        column = list(self.property.remote_side)[0]

        self.manager.association_tables.add(column.table)
        builder = TableBuilder(
            self.manager,
            column.table
        )
        metadata = column.table.metadata
        if builder.parent_table.schema:
            table_name = builder.parent_table.schema + '.' + builder.table_name
        elif metadata.schema:
            table_name = metadata.schema + '.' + builder.table_name
        else:
            table_name = builder.table_name

        if table_name not in metadata.tables:
            self.association_version_table = table = builder()
            self.manager.association_version_tables.add(table)
        else:
            # may have already been created if we visiting the 'other' side of
            # a self-referential many-to-many relationship
            self.association_version_table = metadata.tables[table_name]

    def __call__(self):
        """
        Builds reflected relationship between version classes based on given
        parent object's RelationshipProperty.
        """
        self.local_cls = version_class(self.model)
        self.versioned = False
        try:
            self.remote_cls = version_class(self.property.mapper.class_)
            self.versioned = True
        except (AttributeError, KeyError):
            return
        except ClassNotVersioned:
            self.remote_cls = self.property.mapper.class_

        if (self.property.secondary is not None and
                not self.property.viewonly and
                not self.manager.is_excluded_property(
                    self.model, self.property.key)):
            self.build_association_version_tables()

            # store remote cls to association table column pairs
            self.remote_to_association_column_pairs = []
            for column_pair in self.property.local_remote_pairs:
                if column_pair[0] in self.property.target.c.values():
                    self.remote_to_association_column_pairs.append(column_pair)

        setattr(
            self.local_cls,
            self.property.key,
            self.reflected_relationship
        )



from copy import copy
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import six
import sqlalchemy as sa
from sqlalchemy_utils import identity


class Operation(object):
    INSERT = 0
    UPDATE = 1
    DELETE = 2

    def __init__(self, target, type):
        self.target = target
        self.type = type
        self.processed = False

    def __eq__(self, other):
        return (
            self.target == other.target and
            self.type == other.type
        )

    def __ne__(self, other):
        return not (self == other)


class Operations(object):
    """
    A collection of operations
    """
    def __init__(self):
        self.objects = OrderedDict()

    def format_key(self, target):
        # We cannot use target._sa_instance_state.identity here since object's
        # identity is not yet updated at this phase
        return (target.__class__, identity(target))

    def __contains__(self, target):
        return self.format_key(target) in self.objects

    def __setitem__(self, key, operation):
        self.objects[key] = operation

    def __getitem__(self, key):
        return self.objects[key]

    def __delitem__(self, key):
        del self.objects[key]

    def __bool__(self):
        return bool(self.objects)

    def __nonzero__(self):
        return self.__bool__()

    def __repr__(self):
        return repr(self.objects)

    @property
    def entities(self):
        """
        Return a set of changed versioned entities for given session.

        :param session: SQLAlchemy session object
        """
        return set(key[0] for key, _ in self.iteritems())

    def iteritems(self):
        return six.iteritems(self.objects)

    def items(self):
        return self.objects.items()

    def add(self, operation):
        self[self.format_key(operation.target)] = operation

    def add_insert(self, target):
        if target in self:
            # If the object is deleted and then inserted within the same
            # transaction we are actually dealing with an update.
            self.add(Operation(target, Operation.UPDATE))
        else:
            self.add(Operation(target, Operation.INSERT))

    def add_update(self, target):
        state_copy = copy(sa.inspect(target).committed_state)
        relationships = sa.inspect(target.__class__).relationships
        # Remove all ONETOMANY and MANYTOMANY relationships
        for rel_key, relationship in relationships.items():
            if relationship.direction.name in ['ONETOMANY', 'MANYTOMANY']:
                if rel_key in state_copy:
                    del state_copy[rel_key]

        if state_copy:
            self.add(Operation(target, Operation.UPDATE))

    def add_delete(self, target):
        self.add(Operation(target, Operation.DELETE))



import sqlalchemy as sa
from sqlalchemy_utils import get_column_key


class ColumnReflector(object):
    def __init__(self, manager, parent_table, model=None):
        self.parent_table = parent_table
        self.model = model
        self.manager = manager

    def option(self, name):
        try:
            return self.manager.option(self.model, name)
        except TypeError:
            return self.manager.options[name]

    def reflect_column(self, column):
        """
        Make a copy of parent table column and some alterations to it.

        :param column: SQLAlchemy Column object of parent table
        """
        # Make a copy of the column so that it does not point to wrong table.
        column_copy = column._copy() if hasattr(column, '_copy') else column.copy()
        column_copy.unique = False
        column_copy.onupdate = None
        if column_copy.autoincrement:
            column_copy.autoincrement = False
        if column_copy.name == self.option('transaction_column_name'):
            column_copy.nullable = False

        if not column_copy.primary_key:
            column_copy.nullable = True

        # Find the right column key
        if self.model is not None:
            for key, value in sa.inspect(self.model).columns.items():
                if value is column:
                    column_copy.key = key
        return column_copy

    @property
    def operation_type_column(self):
        """
        Return the operation type column. By default the name of this column
        is 'operation_type'.
        """
        return sa.Column(
            self.option('operation_type_column_name'),
            sa.SmallInteger,
            nullable=False,
            index=True
        )

    @property
    def transaction_column(self):
        """
        Returns transaction column. By default the name of this column is
        'transaction_id'.
        """
        return sa.Column(
            self.option('transaction_column_name'),
            sa.BigInteger,
            primary_key=True,
            index=True,
            autoincrement=False  # This is needed for MySQL
        )

    @property
    def end_transaction_column(self):
        """
        Returns end_transaction column. By default the name of this column is
        'end_transaction_id'.
        """
        return sa.Column(
            self.option('end_transaction_column_name'),
            sa.BigInteger,
            index=True
        )

    @property
    def reflected_parent_columns(self):
        for column in self.parent_table.c:
            if (
                self.model and
                self.manager.is_excluded_column(self.model, column)
            ):
                continue
            reflected_column = self.reflect_column(column)
            yield reflected_column

    def __iter__(self):
        for column in self.reflected_parent_columns:
            yield column

        # Only yield internal version columns if parent model is not using
        # single table inheritance
        if not self.model or not sa.inspect(self.model).single:
            yield self.transaction_column
            if self.option('strategy') == 'validity':
                yield self.end_transaction_column
            yield self.operation_type_column


class TableBuilder(object):
    """
    TableBuilder handles the building of version tables based on parent
    table's structure and versioning configuration options.
    """
    def __init__(
        self,
        versioning_manager,
        parent_table,
        model=None
    ):
        self.manager = versioning_manager
        self.parent_table = parent_table
        self.model = model

    def option(self, name):
        try:
            return self.manager.option(self.model, name)
        except TypeError:
            return self.manager.options[name]

    @property
    def table_name(self):
        """
        Returns the version table name for current parent table.
        """
        return self.option('table_name') % self.parent_table.name

    @property
    def columns(self):
        return list(
            column for column in
            ColumnReflector(self.manager, self.parent_table, self.model)
        )

    def __call__(self, extends=None):
        """
        Builds version table.
        """
        columns = self.columns if extends is None else []
        self.manager.plugins.after_build_version_table_columns(self, columns)
        return sa.schema.Table(
            extends.name if extends is not None else self.table_name,
            self.parent_table.metadata,
            *columns,
            schema=self.parent_table.schema,
            extend_existing=extends is not None
        )



import sqlalchemy as sa

from sqlalchemy_continuum.plugins import PropertyModTrackerPlugin


trigger_sql = """
CREATE TRIGGER {trigger_name}
AFTER INSERT OR UPDATE OR DELETE ON {table_name}
FOR EACH ROW EXECUTE PROCEDURE {procedure_name}()
"""

upsert_cte_sql = """
WITH upsert as
(
    UPDATE {version_table_name}
    SET {update_values}
    WHERE
        {transaction_column} = transaction_id_value
        AND
        {primary_key_criteria}
    RETURNING *
)
INSERT INTO {version_table_name}
({transaction_column}, {operation_type_column}, {column_names})
SELECT
    transaction_id_value,
    {operation_type},
    {insert_values}
WHERE NOT EXISTS (SELECT 1 FROM upsert);
"""

temporary_transaction_sql = """
CREATE TEMP TABLE IF NOT EXISTS {temporary_transaction_table}
({transaction_table_columns})
ON COMMIT DELETE ROWS;
"""

insert_temporary_transaction_sql = """
INSERT INTO {temporary_transaction_table} ({transaction_table_columns})
VALUES ({transaction_values});
"""

temp_transaction_trigger_sql = """
CREATE TRIGGER transaction_trigger
AFTER INSERT ON {transaction_table}
FOR EACH ROW EXECUTE PROCEDURE transaction_temp_table_generator()
"""

procedure_sql = """
CREATE OR REPLACE FUNCTION {procedure_name}() RETURNS TRIGGER AS $$
DECLARE transaction_id_value INT;
BEGIN
    BEGIN
        transaction_id_value = (SELECT id FROM temporary_transaction);
    EXCEPTION WHEN others THEN
        RETURN NEW;
    END;
    IF transaction_id_value IS NULL THEN
        RETURN NEW;
    END IF;

    IF (TG_OP = 'INSERT') THEN
        {after_insert}
        {upsert_insert}
    ELSIF (TG_OP = 'UPDATE') THEN
        IF (hstore(NEW.*) - hstore(OLD.*) - ARRAY[{excluded_columns}]::text[])
            = hstore('')
        THEN
            RETURN NULL;
        END IF;
        {after_update}
        {upsert_update}
    ELSIF (TG_OP = 'DELETE') THEN
        {after_delete}
        {upsert_delete}
    END IF;
    RETURN NEW;
END;
$$
LANGUAGE plpgsql
"""

validity_sql = """
UPDATE {version_table_name}
SET {end_transaction_column} = transaction_id_value
WHERE
    {transaction_column} = (
        SELECT MIN({transaction_column}) FROM {version_table_name}
        WHERE {end_transaction_column} IS NULL AND {primary_key_criteria}
    ) AND
    {primary_key_criteria};
"""


def uses_property_mod_tracking(manager):
    return any(
        isinstance(plugin, PropertyModTrackerPlugin)
        for plugin in manager.plugins
    )


class SQLConstruct(object):
    def __init__(
        self,
        table,
        transaction_column_name,
        operation_type_column_name,
        version_table_name_format,
        excluded_columns=None,
        update_validity_for_tables=None,
        use_property_mod_tracking=False,
        end_transaction_column_name=None,
    ):
        self.update_validity_for_tables = update_validity_for_tables
        self.operation_type_column_name = operation_type_column_name
        self.transaction_column_name = transaction_column_name
        self.end_transaction_column_name = end_transaction_column_name
        self.version_table_name_format = version_table_name_format
        self.use_property_mod_tracking = use_property_mod_tracking
        self.table = table
        self.excluded_columns = excluded_columns
        if update_validity_for_tables is None:
            self.update_validity_for_tables = []
        if self.excluded_columns is None:
            self.excluded_columns = []

    @property
    def table_name(self):
        if self.table.schema:
            return '%s."%s"' % (self.table.schema, self.table.name)
        else:
            return '"' + self.table.name + '"'

    @property
    def transaction_table_name(self):
        if self.table.schema:
            return '%s.transaction' % self.table.schema
        else:
            return 'transaction'

    @property
    def temporary_transaction_table_name(self):
        return 'temporary_transaction'

    @property
    def version_table_name(self):
        version_table_name = self.version_table_name_format % self.table.name
        if self.table.schema:
            version_table_name = '%s.%s' % (
                self.table.schema, version_table_name
            )
        return version_table_name

    @classmethod
    def for_manager(self, manager, cls):
        strategy = manager.option(cls, 'strategy')
        operation_type_column = manager.option(
            cls,
            'operation_type_column_name'
        )
        excluded_columns = [
            c.name for c in sa.inspect(cls).columns
            if manager.is_excluded_column(cls, c)
        ]
        return self(
            update_validity_for_tables=(
                sa.inspect(cls).tables if strategy == 'validity' else []
            ),
            version_table_name_format=manager.option(cls, 'table_name'),
            operation_type_column_name=operation_type_column,
            transaction_column_name=manager.option(
                cls, 'transaction_column_name'
            ),
            end_transaction_column_name=manager.option(
                cls, 'end_transaction_column_name'
            ),
            use_property_mod_tracking=uses_property_mod_tracking(manager),
            excluded_columns=excluded_columns,
            table=cls.__table__
        )

    @property
    def columns(self):
        return [c for c in self.table.c if c.name not in self.excluded_columns]

    @property
    def columns_without_pks(self):
        return [c for c in self.columns if not c.primary_key]

    @property
    def pk_columns(self):
        return [c for c in self.columns if c.primary_key]

    def copy_args(self):
        return dict(
            (k, v) for k, v in self.__dict__.items() if not k.startswith('__')
        )


class UpsertSQL(SQLConstruct):
    builders = {
        'update_values': ', ',
        'insert_values': ', ',
        'column_names': ', ',
        'primary_key_criteria': ' AND ',
    }

    def __init__(self, *args, **kwargs):
        SQLConstruct.__init__(self, *args, **kwargs)

        for key in self.builders:
            setattr(self, key, getattr(self, 'build_%s' % key)())

    def build_column_names(self):
        column_names = ['"%s"' % c.name for c in self.columns]
        if self.use_property_mod_tracking:
            column_names += [
                '%s_mod' % c.name for c in self.columns_without_pks
            ]
        return column_names

    def build_primary_key_criteria(self):
        return [
            '"{name}" = NEW."{name}"'.format(name=c.name)
            for c in self.columns if c.primary_key
        ]

    def build_update_values(self):
        parent_columns = [
            '"{name}" = NEW."{name}"'.format(name=c.name)
            for c in self.columns
        ]
        mod_columns = []
        if self.use_property_mod_tracking:
            mod_columns = [
                '{0}_mod = {0}_mod OR OLD."{0}" IS DISTINCT FROM NEW."{0}"'
                .format(c.name)
                for c in self.columns_without_pks
            ]

        return (
            ['%s = 1' % self.operation_type_column_name] +
            parent_columns +
            mod_columns
        )

    def build_insert_values(self):
        values = self.build_values()
        if self.use_property_mod_tracking:
            values += self.build_mod_tracking_values()
        return values

    def build_values(self):
        return ['NEW."%s"' % c.name for c in self.columns]

    def build_mod_tracking_values(self):
        return []

    def __str__(self):
        params = dict(
            version_table_name=self.version_table_name,
            transaction_column=self.transaction_column_name,
            operation_type=self.operation_type,
            operation_type_column=self.operation_type_column_name,
            transaction_table_name=self.transaction_table_name,
        )
        for key, join_operator in self.builders.items():
            params[key] = join_operator.join(getattr(self, key))

        sql = upsert_cte_sql.format(**params)
        return sql


class DeleteUpsertSQL(UpsertSQL):
    operation_type = 2

    def build_primary_key_criteria(self):
        return [
            '"{name}" = OLD."{name}"'.format(name=c.name)
            for c in self.pk_columns
        ]

    def build_mod_tracking_values(self):
        return ['True'] * len(self.columns_without_pks)

    def build_update_values(self):
        return [
            '"{name}" = OLD."{name}"'.format(name=c.name)
            for c in self.columns
        ]

    def build_values(self):
        return ['OLD."%s"' % c.name for c in self.columns]


class InsertUpsertSQL(UpsertSQL):
    operation_type = 0

    def build_mod_tracking_values(self):
        return ['True'] * len(self.columns_without_pks)


class UpdateUpsertSQL(UpsertSQL):
    operation_type = 1

    def build_mod_tracking_values(self):
        return [
            'OLD."{0}" IS DISTINCT FROM NEW."{0}"'
            .format(c.name) for c in self.columns_without_pks
        ]


class ValiditySQL(SQLConstruct):
    @property
    def primary_key_criteria(self):
        return ' AND '.join(
            '"{name}" = NEW."{name}"'.format(name=c.name)
            for c in self.pk_columns
        )

    def __str__(self):
        params = dict(
            version_table_name=self.version_table_name,
            transaction_table_name=self.transaction_table_name,
            transaction_column=self.transaction_column_name,
            end_transaction_column=self.end_transaction_column_name,
            primary_key_criteria=self.primary_key_criteria
        )
        return validity_sql.format(**params)


class InsertValiditySQL(ValiditySQL):
    pass


class UpdateValiditySQL(ValiditySQL):
    pass


class DeleteValiditySQL(ValiditySQL):
    @property
    def primary_key_criteria(self):
        return ' AND '.join(
            '{name} = OLD."{name}"'.format(name=c.name)
            for c in self.pk_columns
        )


def get_validity_sql(class_, tables, params):
    params = params.copy()
    del params['table']
    return ''.join(str(class_(table, **params)) for table in tables)


class CreateTriggerSQL(SQLConstruct):
    def __str__(self):
        return trigger_sql.format(
            trigger_name='%s_trigger' % self.table.name,
            table_name=self.table_name,
            procedure_name='%s_audit' % self.table.name
        )


class TransactionSQLConstruct(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class CreateTemporaryTransactionTableSQL(TransactionSQLConstruct):
    table_name = 'temporary_transaction'

    def __str__(self):
        return temporary_transaction_sql.format(
            temporary_transaction_table=self.table_name,
            transaction_table_columns='id BIGINT, PRIMARY KEY(id)'
        )


class InsertTemporaryTransactionSQL(TransactionSQLConstruct):
    table_name = 'temporary_transaction'
    transaction_values = 'transaction_id_value'

    def __str__(self):
        return insert_temporary_transaction_sql.format(
            temporary_transaction_table=self.table_name,
            transaction_table_columns='id',
            transaction_values=self.transaction_values
        )


class CreateTriggerFunctionSQL(SQLConstruct):
    def __str__(self):
        args = self.copy_args()
        tables = self.update_validity_for_tables
        after_insert = get_validity_sql(InsertValiditySQL, tables, args)
        after_update = get_validity_sql(UpdateValiditySQL, tables, args)
        after_delete = get_validity_sql(DeleteValiditySQL, tables, args)

        sql = procedure_sql.format(
            procedure_name='%s_audit' % self.table.name,
            excluded_columns=', '.join(
                "'%s'" % c for c in self.excluded_columns
            ),
            transaction_table_name=self.transaction_table_name,
            after_insert=after_insert,
            after_update=after_update,
            after_delete=after_delete,
            temporary_transaction_sql=(
                CreateTemporaryTransactionTableSQL()
            ),
            insert_temporary_transaction_sql=(
                InsertTemporaryTransactionSQL()
            ),
            upsert_insert=InsertUpsertSQL(**args),
            upsert_update=UpdateUpsertSQL(**args),
            upsert_delete=DeleteUpsertSQL(**args)
        )
        return sql


class TransactionTriggerSQL(object):
    def __init__(self, tx_class):
        self.table = tx_class.__table__

    @property
    def transaction_table_name(self):
        if self.table.schema:
            return '%s.transaction' % self.table.schema
        else:
            return 'transaction'

    def __str__(self):
        return temp_transaction_trigger_sql.format(
            transaction_table=self.transaction_table_name
        )


def create_versioning_trigger_listeners(manager, cls):
    sa.event.listen(
        cls.__table__,
        'after_create',
        sa.schema.DDL(str(CreateTriggerFunctionSQL.for_manager(manager, cls)))
    )
    sa.event.listen(
        cls.__table__,
        'after_create',
        sa.schema.DDL(str(CreateTriggerSQL.for_manager(manager, cls)))
    )
    sa.event.listen(
        cls.__table__,
        'after_drop',
        sa.schema.DDL(
            'DROP FUNCTION IF EXISTS %s()' %
            '%s_audit' % cls.__table__.name,
        )
    )


def sync_trigger(conn, table_name, **kwargs):
    """
    Synchronizes versioning trigger for given table with given connection.

    ::


        sync_trigger(conn, 'my_table')


    :param conn: SQLAlchemy connection object
    :param table_name: Name of the table to synchronize versioning trigger for
    :params **kwargs: kwargs to pass to create_trigger

    .. versionadded: 1.1.0
    """
    meta = sa.MetaData()
    version_table = sa.Table(
        table_name,
        meta,
        autoload=True,
        autoload_with=conn
    )
    parent_table = sa.Table(
        table_name[0:-len('_version')],
        meta,
        autoload=True,
        autoload_with=conn
    )
    excluded_columns = (
        set(c.name for c in parent_table.c) -
        set(c.name for c in version_table.c if not c.name.endswith('_mod'))
    )
    drop_trigger(conn, parent_table.name)
    create_trigger(conn, table=parent_table, excluded_columns=excluded_columns, **kwargs)


def create_trigger(
    conn,
    table,
    transaction_column_name='transaction_id',
    operation_type_column_name='operation_type',
    version_table_name_format='%s_version',
    excluded_columns=None,
    use_property_mod_tracking=True,
    end_transaction_column_name=None,
):
    params = dict(
        table=table,
        update_validity_for_tables=[],
        transaction_column_name=transaction_column_name,
        operation_type_column_name=operation_type_column_name,
        version_table_name_format=version_table_name_format,
        excluded_columns=excluded_columns,
        use_property_mod_tracking=use_property_mod_tracking,
        end_transaction_column_name=end_transaction_column_name,
    )
    conn.execute(str(CreateTriggerFunctionSQL(**params)))
    conn.execute(str(CreateTriggerSQL(**params)))


def drop_trigger(conn, table_name):
    conn.execute(
        'DROP TRIGGER IF EXISTS %s_trigger ON "%s"' % (
            table_name,
            table_name
        )
    )
    conn.execute('DROP FUNCTION IF EXISTS %s_audit()' % table_name)



import operator
import sqlalchemy as sa
from sqlalchemy_utils import get_primary_keys, identity
from .utils import tx_column_name, end_tx_column_name


def parent_identity(obj_or_class):
    return tuple(
        getattr(obj_or_class, column_key)
        for column_key in get_primary_keys(obj_or_class).keys()
        if column_key != tx_column_name(obj_or_class)
    )


def eqmap(callback, iterable):
    for a, b in zip(*map(callback, iterable)):
        yield a == b


def parent_criteria(obj, class_=None):
    if class_ is None:
        class_ = obj.__class__
    return eqmap(parent_identity, (class_, obj))


class VersionObjectFetcher(object):
    def __init__(self, manager):
        self.manager = manager

    def previous(self, obj):
        """
        Returns the previous version relative to this version in the version
        history. If current version is the first version this method returns
        None.
        """
        return self.previous_query(obj).first()

    def index(self, obj):
        """
        Return the index of this version in the version history.
        """
        session = sa.orm.object_session(obj)
        return session.execute(self._index_query(obj)).fetchone()[0]

    def next(self, obj):
        """
        Returns the next version relative to this version in the version
        history. If current version is the last version this method returns
        None.
        """
        return self.next_query(obj).first()

    def _transaction_id_subquery(self, obj, next_or_prev='next', alias=None):
        if next_or_prev == 'next':
            op = operator.gt
            func = sa.func.min
        else:
            op = operator.lt
            func = sa.func.max

        if alias is None:
            alias = sa.orm.aliased(obj.__class__)
            table = alias.__table__
            if hasattr(alias, 'c'):
                attrs = alias.c
            else:
                attrs = alias
        else:
            table = alias.original
            attrs = alias.c
        query = (
            sa.select(
                [func(
                    getattr(attrs, tx_column_name(obj))
                )],
                from_obj=[table]
            )
            .where(
                sa.and_(
                    op(
                        getattr(attrs, tx_column_name(obj)),
                        getattr(obj, tx_column_name(obj))
                    ),
                    *[
                        getattr(attrs, pk) == getattr(obj, pk)
                        for pk in get_primary_keys(obj.__class__)
                        if pk != tx_column_name(obj)
                    ]
                )
            )
            .correlate(table)
        )
        try:
            return query.scalar_subquery()
        except AttributeError:  # SQLAlchemy < 1.4
            return query.as_scalar()

    def _next_prev_query(self, obj, next_or_prev='next'):
        session = sa.orm.object_session(obj)

        subquery = self._transaction_id_subquery(
            obj, next_or_prev=next_or_prev
        )
        try:
            subquery = subquery.scalar_subquery()
        except AttributeError:  # SQLAlchemy < 1.4
            subquery = subquery.as_scalar()

        return (
            session.query(obj.__class__)
            .filter(
                sa.and_(
                    getattr(
                        obj.__class__,
                        tx_column_name(obj)
                    ) == subquery,
                    *parent_criteria(obj)
                )
            )
        )

    def _index_query(self, obj):
        """
        Returns the query needed for fetching the index of this record relative
        to version history.
        """
        alias = sa.orm.aliased(obj.__class__)

        subquery = (
            sa.select([sa.func.count('1')], from_obj=[alias.__table__])
            .where(
                getattr(alias, tx_column_name(obj))
                <
                getattr(obj, tx_column_name(obj))
            )
            .correlate(alias.__table__)
            .label('position')
        )
        query = (
            sa.select([subquery], from_obj=[obj.__table__])
            .where(
                sa.and_(*eqmap(identity, (obj.__class__, obj)))
            )
            .order_by(
                getattr(obj.__class__, tx_column_name(obj))
            )
        )
        return query


class SubqueryFetcher(VersionObjectFetcher):
    def previous_query(self, obj):
        """
        Returns the query that fetches the previous version relative to this
        version in the version history.
        """
        return self._next_prev_query(obj, 'previous')

    def next_query(self, obj):
        """
        Returns the query that fetches the next version relative to this
        version in the version history.
        """
        return self._next_prev_query(obj, 'next')


class ValidityFetcher(VersionObjectFetcher):
    def next_query(self, obj):
        """
        Returns the query that fetches the next version relative to this
        version in the version history.
        """
        session = sa.orm.object_session(obj)

        return (
            session.query(obj.__class__)
            .filter(
                sa.and_(
                    getattr(obj.__class__, tx_column_name(obj))
                    ==
                    getattr(obj, end_tx_column_name(obj)),
                    *parent_criteria(obj)
                )
            )
        )

    def previous_query(self, obj):
        """
        Returns the query that fetches the previous version relative to this
        version in the version history.
        """
        session = sa.orm.object_session(obj)

        return (
            session.query(obj.__class__)
            .filter(
                sa.and_(
                    getattr(obj.__class__, end_tx_column_name(obj))
                    ==
                    getattr(obj, tx_column_name(obj)),
                    *parent_criteria(obj)
                )
            )
        )



class ModelFactory(object):
    model_name = None

    def __call__(self, manager):
        """
        Create model class but only if it doesn't already exist
        in declarative model registry.
        """
        Base = manager.declarative_base
        try:
            registry = Base.registry._class_registry
        except AttributeError:  # SQLAlchemy < 1.4
            registry = Base._decl_class_registry
        if self.model_name not in registry:
            return self.create_class(manager)
        return registry[self.model_name]



from datetime import datetime
from functools import partial

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import six
import sqlalchemy as sa
from sqlalchemy.ext.compiler import compiles

from .dialects.postgresql import (
    CreateTemporaryTransactionTableSQL,
    InsertTemporaryTransactionSQL,
    TransactionTriggerSQL
)
from .exc import ImproperlyConfigured
from .factory import ModelFactory


@compiles(sa.types.BigInteger, 'sqlite')
def compile_big_integer(element, compiler, **kw):
    return 'INTEGER'


class NoChangesAttribute(Exception):
    pass


class TransactionBase(object):
    issued_at = sa.Column(sa.DateTime, default=datetime.utcnow)

    @property
    def entity_names(self):
        """
        Return a list of entity names that changed during this transaction.
        Raises a NoChangesAttribute exception if the 'changes' column does
        not exist, most likely because TransactionChangesPlugin is not enabled.
        """
        if hasattr(self, 'changes'):
          return [changes.entity_name for changes in self.changes]
        else:
          raise NoChangesAttribute()

    @property
    def changed_entities(self):
        """
        Return all changed entities for this transaction log entry.

        Entities are returned as a dict where keys are entity classes and
        values lists of entitites that changed in this transaction.
        """
        manager = self.__versioning_manager__
        tuples = set(manager.version_class_map.items())
        entities = {}

        session = sa.orm.object_session(self)

        for class_, version_class in tuples:
            try:
                if class_.__name__ not in self.entity_names:
                    continue
            except NoChangesAttribute:
                pass

            tx_column = manager.option(class_, 'transaction_column_name')

            entities[version_class] = (
                session
                .query(version_class)
                .filter(getattr(version_class, tx_column) == self.id)
            ).all()
        return entities


procedure_sql = """
CREATE OR REPLACE FUNCTION transaction_temp_table_generator()
RETURNS TRIGGER AS $$
BEGIN
    {temporary_transaction_sql}
    INSERT INTO temporary_transaction (id) VALUES (NEW.id);
    RETURN NEW;
END;
$$
LANGUAGE plpgsql
"""


def create_triggers(cls):
    sa.event.listen(
        cls.__table__,
        'after_create',
        sa.schema.DDL(
            procedure_sql.format(
                temporary_transaction_sql=CreateTemporaryTransactionTableSQL(),
                insert_temporary_transaction_sql=(
                    InsertTemporaryTransactionSQL(
                        transaction_id_values='NEW.id'
                    )
                ),
            )
        )
    )
    sa.event.listen(
        cls.__table__,
        'after_create',
        sa.schema.DDL(str(TransactionTriggerSQL(cls)))
    )
    sa.event.listen(
        cls.__table__,
        'after_drop',
        sa.schema.DDL(
            'DROP FUNCTION IF EXISTS transaction_temp_table_generator()'
        )
    )


class TransactionFactory(ModelFactory):
    model_name = 'Transaction'

    def __init__(self, remote_addr=True):
        self.remote_addr = remote_addr

    def create_class(self, manager):
        """
        Create Transaction class.
        """
        class Transaction(
            manager.declarative_base,
            TransactionBase
        ):
            __tablename__ = 'transaction'
            __versioning_manager__ = manager

            id = sa.Column(
                sa.types.BigInteger,
                sa.schema.Sequence('transaction_id_seq'),
                primary_key=True,
                autoincrement=True
            )

            if self.remote_addr:
                remote_addr = sa.Column(sa.String(50))

            if manager.user_cls:
                user_cls = manager.user_cls
                Base = manager.declarative_base
                try:
                    registry = Base.registry._class_registry
                except AttributeError:  # SQLAlchemy < 1.4
                    registry = Base._decl_class_registry

                if isinstance(user_cls, six.string_types):
                    try:
                        user_cls = registry[user_cls]
                    except KeyError:
                        raise ImproperlyConfigured(
                            'Could not build relationship between Transaction'
                            ' and %s. %s was not found in declarative class '
                            'registry. Either configure VersioningManager to '
                            'use different user class or disable this '
                            'relationship ' % (user_cls, user_cls)
                        )

                user_id = sa.Column(
                    sa.inspect(user_cls).primary_key[0].type,
                    sa.ForeignKey(sa.inspect(user_cls).primary_key[0]),
                    index=True
                )

                user = sa.orm.relationship(user_cls)

            def __repr__(self):
                fields = ['id', 'issued_at', 'user']
                field_values = OrderedDict(
                    (field, getattr(self, field))
                    for field in fields
                    if hasattr(self, field)
                )
                return '<Transaction %s>' % ', '.join(
                    (
                        '%s=%r' % (field, value)
                        if not isinstance(value, six.integer_types)
                        # We want the following line to ensure that longs get
                        # shown without the ugly L suffix on python 2.x
                        # versions
                        else '%s=%d' % (field, value)
                        for field, value in field_values.items()
                    )
                )

        if manager.options['native_versioning']:
            create_triggers(Transaction)
        return Transaction



import sqlalchemy as sa


def get_end_tx_column_query(
    table,
    end_tx_column_name='end_transaction_id',
    tx_column_name='transaction_id'
):

    v1 = sa.alias(table, name='v')
    v2 = sa.alias(table, name='v2')
    v3 = sa.alias(table, name='v3')

    primary_keys = [c.name for c in table.c if c.primary_key]

    tx_criterion = sa.select(
        [sa.func.min(getattr(v3.c, tx_column_name))]
    ).where(
        sa.and_(
            getattr(v3.c, tx_column_name) > getattr(v1.c, tx_column_name),
            *[
                getattr(v3.c, pk) == getattr(v1.c, pk)
                for pk in primary_keys
                if pk != tx_column_name
            ]
        )
    )
    try:
        tx_criterion = tx_criterion.scalar_subquery()
    except AttributeError:  # SQLAlchemy < 1.4
        tx_criterion = tx_criterion.as_scalar()
    return sa.select(
        columns=[
            getattr(v1.c, column)
            for column in primary_keys
        ] + [
            getattr(v2.c, tx_column_name).label(end_tx_column_name)
        ],
        from_obj=v1.outerjoin(
            v2,
            sa.and_(
                getattr(v2.c, tx_column_name) ==
                tx_criterion
            )
        )
    ).order_by(getattr(v1.c, tx_column_name))


def update_end_tx_column(
    table,
    end_tx_column_name='end_transaction_id',
    tx_column_name='transaction_id',
    conn=None
):
    """
    Calculates end transaction columns and updates the version table with the
    calculated values. This function can be used for migrating between subquery
    versioning strategy and validity versioning strategy.

    :param table: SQLAlchemy table object
    :param end_tx_column_name: Name of the end transaction column
    :param tx_column_name: Transaction column name
    :param conn:
        Either SQLAlchemy Connection, Engine, Session or Alembic
        Operations object. Basically this should be an object that can execute
        the queries needed to update the end transaction column values.

        If no object is given then this function tries to use alembic.op for
        executing the queries.
    """
    if conn is None:
        from alembic import op

        conn = op.get_bind()

    query = get_end_tx_column_query(
        table,
        end_tx_column_name=end_tx_column_name,
        tx_column_name=tx_column_name
    )
    stmt = conn.execute(query)
    primary_keys = [c.name for c in table.c if c.primary_key]
    for row in stmt:
        if row[end_tx_column_name]:
            criteria = [
                getattr(table.c, pk) == row[pk]
                for pk in primary_keys
            ]

            update_stmt = (
                table.update()
                .where(sa.and_(*criteria))
                .values({end_tx_column_name: row[end_tx_column_name]})
            )
            conn.execute(update_stmt)


def get_property_mod_flags_query(
    table,
    tracked_columns,
    mod_suffix='_mod',
    end_tx_column_name='end_transaction_id',
    tx_column_name='transaction_id',
):
    v1 = sa.alias(table, name='v')
    v2 = sa.alias(table, name='v2')
    primary_keys = [c.name for c in table.c if c.primary_key]

    return sa.select(
        columns=[
            getattr(v1.c, column)
            for column in primary_keys
        ] + [
            (sa.or_(
                getattr(v1.c, column) != getattr(v2.c, column),
                getattr(v2.c, tx_column_name).is_(None)
            )).label(column + mod_suffix)
            for column in tracked_columns
        ],
        from_obj=v1.outerjoin(
            v2,
            sa.and_(
                getattr(v2.c, end_tx_column_name) ==
                getattr(v1.c, tx_column_name),
                *[
                    getattr(v2.c, pk) == getattr(v1.c, pk)
                    for pk in primary_keys
                    if pk != tx_column_name
                ]
            )
        )
    ).order_by(getattr(v1.c, tx_column_name))


def update_property_mod_flags(
    table,
    tracked_columns,
    mod_suffix='_mod',
    end_tx_column_name='end_transaction_id',
    tx_column_name='transaction_id',
    conn=None
):
    """
    Update property modification flags for given table and given columns. This
    function can be used for migrating an existing schema to use property mod
    flags (provided by PropertyModTracker plugin).

    :param table: SQLAlchemy table object
    :param mod_suffix: Modification tracking columns suffix
    :param end_tx_column_name: Name of the end transaction column
    :param tx_column_name: Transaction column name
    :param conn:
        Either SQLAlchemy Connection, Engine, Session or Alembic
        Operations object. Basically this should be an object that can execute
        the queries needed to update the property modification flags.

        If no object is given then this function tries to use alembic.op for
        executing the queries.
    """
    if conn is None:
        from alembic import op

        conn = op.get_bind()

    query = get_property_mod_flags_query(
        table,
        tracked_columns,
        mod_suffix=mod_suffix,
        end_tx_column_name=end_tx_column_name,
        tx_column_name=tx_column_name,
    )
    stmt = conn.execute(query)

    primary_keys = [c.name for c in table.c if c.primary_key]
    for row in stmt:
        values = dict([
            (column + mod_suffix, row[column + mod_suffix])
            for column in tracked_columns
            if row[column + mod_suffix]
        ])
        if values:
            criteria = [
                getattr(table.c, pk) == row[pk] for pk in primary_keys
            ]
            query = table.update().where(sa.and_(*criteria)).values(values)
            conn.execute(query)



import sqlalchemy as sa
from sqlalchemy.sql.expression import bindparam

from .utils import version_table


class VersionExpressionReflector(sa.sql.visitors.ReplacingCloningVisitor):
    def __init__(self, parent, relationship):
        self.parent = parent
        self.relationship = relationship

    def replace(self, column):
        if not isinstance(column, sa.Column):
            return
        try:
            table = version_table(column.table)
        except KeyError:
            reflected_column = column
        else:
            reflected_column = table.c[column.name]
            if (
                column in self.relationship.local_columns and
                table == self.parent.__table__
            ):
                reflected_column = bindparam(
                    column.key,
                    getattr(self.parent, column.key)
                )

        return reflected_column

    def __call__(self, expr):
        return self.traverse(expr)



from copy import copy

import sqlalchemy as sa
from sqlalchemy_utils import get_primary_keys, identity
from .operation import Operations
from .utils import (
    end_tx_column_name,
    version_class,
    is_session_modified,
    tx_column_name,
    versioned_column_properties
)


class UnitOfWork(object):
    def __init__(self, manager):
        self.manager = manager
        self.reset()

    def reset(self, session=None):
        """
        Reset the internal state of this UnitOfWork object. Normally this is
        called after transaction has been committed or rolled back.
        """
        self.version_session = None
        self.current_transaction = None
        self.operations = Operations()
        self.pending_statements = []
        self.version_objs = {}

    def is_modified(self, session):
        """
        Return whether or not given session has been modified. Session has been
        modified if any versioned property of any version object in given
        session has been modified or if any of the plugins returns that
        session has been modified.

        :param session: SQLAlchemy session object
        """
        return (
            is_session_modified(session) or
            any(self.manager.plugins.is_session_modified(session))
        )

    def process_before_flush(self, session):
        """
        Before flush processor for given session.

        This method creates a version session which is later on used for the
        creation of version objects. It also creates Transaction object for the
        current transaction and invokes before_flush template method on all
        plugins.

        If the given session had no relevant modifications regarding versioned
        objects this method does nothing.

        :param session: SQLAlchemy session object
        """
        if session == self.version_session:
            return

        if not self.is_modified(session):
            return

        if not self.version_session:
            self.version_session = sa.orm.session.Session(
                bind=session.connection()
            )

        if not self.current_transaction:
            self.create_transaction(session)

        self.manager.plugins.before_flush(self, session)

    def process_after_flush(self, session):
        """
        After flush processor for given session.

        Creates version objects for all modified versioned parent objects that
        were affected during the flush phase.

        :param session: SQLAlchemy session object
        """
        if session == self.version_session:
            return

        if not self.current_transaction:
            return

        if not self.version_session:
            self.version_session = sa.orm.session.Session(
                bind=session.connection()
            )

        self.make_versions(session)

    def transaction_args(self, session):
        args = {}
        for plugin in self.manager.plugins:
            args.update(plugin.transaction_args(self, session))
        return args

    def create_transaction(self, session):
        """
        Create transaction object for given SQLAlchemy session.

        :param session: SQLAlchemy session object
        """
        args = self.transaction_args(session)

        Transaction = self.manager.transaction_cls
        self.current_transaction = Transaction()

        for key, value in args.items():
            setattr(self.current_transaction, key, value)
        if not self.version_session:
            self.version_session = sa.orm.session.Session(
                bind=session.connection()
            )
        self.version_session.add(self.current_transaction)
        self.version_session.flush()
        self.version_session.expunge(self.current_transaction)
        session.add(self.current_transaction)
        return self.current_transaction

    def get_or_create_version_object(self, target):
        """
        Return version object for given parent object. If no version object
        exists for given parent object, create one.

        :param target: Parent object to create the version object for
        """
        version_cls = version_class(target.__class__)
        version_id = identity(target) + (self.current_transaction.id, )
        version_key = (version_cls, version_id)

        if version_key not in self.version_objs:
            version_obj = version_cls()
            self.version_objs[version_key] = version_obj
            self.version_session.add(version_obj)
            tx_column = self.manager.option(
                target,
                'transaction_column_name'
            )
            setattr(
                version_obj,
                tx_column,
                self.current_transaction.id
            )
            return version_obj
        else:
            return self.version_objs[version_key]

    def process_operation(self, operation):
        """
        Process given operation object. The operation processing has x stages:

        1. Get or create a version object for given parent object
        2. Assign the operation type for this object
        3. Invoke listeners
        4. Update version validity in case validity strategy is used
        5. Mark operation as processed

        :param operation: Operation object
        """
        target = operation.target
        version_obj = self.get_or_create_version_object(target)
        version_obj.operation_type = operation.type
        self.assign_attributes(target, version_obj)

        self.manager.plugins.after_create_version_object(
            self, target, version_obj
        )
        if self.manager.option(target, 'strategy') == 'validity':
            self.update_version_validity(
                target,
                version_obj
            )
        operation.processed = True

    def create_version_objects(self, session):
        """
        Create version objects for given session based on operations collected
        by insert, update and deleted trackers.

        :param session: SQLAlchemy session object
        """
        if (
            not self.manager.options['versioning'] or
            self.manager.options['native_versioning']
        ):
            return

        for key, operation in copy(self.operations).items():
            if operation.processed:
                continue

            if not self.current_transaction:
                raise Exception(
                    'Current transaction not available.'
                )
            self.process_operation(operation)

        self.version_session.flush()

    def version_validity_subquery(self, parent, version_obj, alias=None):
        """
        Return the subquery needed by :func:`update_version_validity`.

        This method is only used when using 'validity' versioning strategy.

        :param parent: SQLAlchemy declarative parent object
        :parem version_obj: SQLAlchemy declarative version object

        .. seealso:: :func:`update_version_validity`
        """
        fetcher = self.manager.fetcher(parent)
        session = sa.orm.object_session(version_obj)

        subquery = fetcher._transaction_id_subquery(
            version_obj,
            next_or_prev='prev',
            alias=alias
        )
        if session.connection().engine.dialect.name == 'mysql':
            return sa.select(
                [sa.text('max_1')],
                from_obj=[
                    sa.sql.expression.alias(subquery.subquery() if hasattr(subquery, 'subquery') else subquery, name='subquery')
                ]
            )
        return subquery

    def update_version_validity(self, parent, version_obj):
        """
        Updates previous version object end_transaction_id based on given
        parent object and newly created version object.

        This method is only used when using 'validity' versioning strategy.

        :param parent: SQLAlchemy declarative parent object
        :parem version_obj: SQLAlchemy declarative version object

        .. seealso:: :func:`version_validity_subquery`
        """
        session = sa.orm.object_session(version_obj)

        for class_ in version_obj.__class__.__mro__:
            if class_ in self.manager.parent_class_map:

                subquery = self.version_validity_subquery(
                    parent,
                    version_obj,
                    alias=sa.orm.aliased(class_.__table__)
                )
                try:
                    subquery = subquery.scalar_subquery()
                except AttributeError:  # SQLAlchemy < 1.4
                    subquery = subquery.as_scalar()

                query = (
                    session.query(class_.__table__)
                    .filter(
                        sa.and_(
                            getattr(
                                class_,
                                tx_column_name(version_obj)
                            ) == subquery,
                            *[
                                getattr(version_obj, pk) ==
                                getattr(class_.__table__.c, pk)
                                for pk in get_primary_keys(class_)
                                if pk != tx_column_name(class_)
                            ]
                        )
                    )
                )
                query.update(
                    {
                        end_tx_column_name(version_obj):
                        self.current_transaction.id
                    },
                    synchronize_session=False
                )

    def create_association_versions(self, session):
        """
        Creates association table version records for given session.

        :param session: SQLAlchemy session object
        """
        statements = copy(self.pending_statements)
        for stmt in statements:
            stmt = stmt.values(
                **{
                    self.manager.options['transaction_column_name']:
                    self.current_transaction.id
                }
            )
            session.execute(stmt)
        self.pending_statements = []

    def make_versions(self, session):
        """
        Create transaction, transaction changes records, version objects.

        :param session: SQLAlchemy session object
        """
        if not self.manager.options['versioning']:
            return

        if self.pending_statements:
            self.create_association_versions(session)

        if self.operations:
            self.manager.plugins.before_create_version_objects(self, session)
            self.create_version_objects(session)
            self.manager.plugins.after_create_version_objects(self, session)

    @property
    def has_changes(self):
        """
        Return whether or not this unit of work has changes.
        """
        return self.operations or self.pending_statements

    def assign_attributes(self, parent_obj, version_obj):
        """
        Assign attributes values from parent object to version object.

        :param parent_obj:
            Parent object to get the attribute values from
        :param version_obj:
            Version object to assign the attribute values to
        """
        for prop in versioned_column_properties(parent_obj):
            try:
                value = getattr(parent_obj, prop.key)
            except sa.orm.exc.ObjectDeletedError:
                value = None
            setattr(version_obj, prop.key, value)



class Plugin(object):
    def is_session_modified(self, session):
        return False

    def after_build_tx_class(self, manager):
        pass

    def after_build_models(self, manager):
        pass

    def after_build_version_table_columns(self, table_builder, columns):
        pass

    def before_flush(self, uow, session):
        pass

    def before_create_version_objects(self, uow, session):
        pass

    def after_create_version_objects(self, uow, session):
        pass

    def after_create_version_object(self, uow, parent_obj, version_obj):
        pass

    def transaction_args(self, uow, session):
        return {}

    def after_version_class_built(self, parent_cls, version_cls):
        pass

    def after_construct_changeset(self, version_obj, changeset):
        pass


class PluginCollection(object):
    def __init__(self, plugins=None):
        if plugins is None:
            plugins = []
        if isinstance(plugins, self.__class__):
            self.plugins = plugins.plugins
        else:
            self.plugins = plugins

    def __iter__(self):
        for plugin in self.plugins:
            yield plugin

    def __len__(self):
        return len(self.plugins)

    def __repr__(self):
        return '<%s [%s]>' % (
            self.__class__.__name__,
            ', '.join(map(repr, self.plugins))
        )

    def __getitem__(self, index):
        return self.plugins[index]

    def __setitem__(self, index, element):
        self.plugins[index] = element

    def __delitem__(self, index):
        del self.plugins[index]

    def __getattr__(self, attr):
        def wrapper(*args, **kwargs):
            return [
                getattr(plugin, attr)(*args, **kwargs)
                for plugin in self.plugins
            ]
        return wrapper

    def append(self, el):
        self.plugins.append(el)



from copy import copy
from inspect import getmro
from functools import wraps

import sqlalchemy as sa
from sqlalchemy_utils.functions import get_declarative_base

from .dialects.postgresql import create_versioning_trigger_listeners
from .model_builder import ModelBuilder
from .relationship_builder import RelationshipBuilder
from .table_builder import TableBuilder


def prevent_reentry(handler):
    in_handler = False
    @wraps(handler)
    def check_reentry(*args, **kwargs):
        nonlocal in_handler
        if in_handler:
            return
        in_handler = True
        handler(*args, **kwargs)
        in_handler = False
    return check_reentry

class Builder(object):
    def build_triggers(self):
        """
        Build native database versioning triggers for all versioned models that
        were collected during class instrumentation process.
        """
        processed_tables = set()
        for cls in self.manager.pending_classes:
            if not self.manager.option(cls, 'versioning'):
                continue

            if self.manager.option(cls, 'native_versioning'):
                cls.__versioning_manager__ = self.manager

                if cls.__table__ not in processed_tables:
                    create_versioning_trigger_listeners(self.manager, cls)
                    processed_tables.add(cls.__table__)

    def build_tables(self):
        """
        Build tables for version models based on classes that were collected
        during class instrumentation process.
        """
        for cls in self.manager.pending_classes:
            if not self.manager.option(cls, 'versioning'):
                continue

            if self.manager.option(cls, 'create_tables'):
                inherited_table = None
                for class_ in self.manager.tables:
                    if (issubclass(cls, class_) and
                            cls.__table__ == class_.__table__):
                        inherited_table = self.manager.tables[class_]
                        break

                builder = TableBuilder(
                    self.manager,
                    cls.__table__,
                    model=cls
                )
                if inherited_table is not None:
                    self.manager.tables[class_] = builder(inherited_table)
                else:
                    table = builder()
                    self.manager.tables[cls] = table

    def closest_matching_table(self, model):
        """
        Returns the closest matching table from the generated tables dictionary
        for given model. First tries to fetch an exact match for given model.
        If no table was found then tries to match given model as a subclass.

        :param model: SQLAlchemy declarative model class.
        """
        if model in self.manager.tables:
            return self.manager.tables[model]
        subclasses = [cls for cls in self.manager.tables if issubclass(model, cls)]
        ordered_subclasses = [cls for cls in getmro(model) if cls in subclasses]
        return self.manager.tables[ordered_subclasses[0]] if ordered_subclasses else None


    def build_models(self):
        """
        Build declarative version models based on classes that were collected
        during class instrumentation process.
        """
        if self.manager.pending_classes:
            for cls in self.manager.pending_classes:
                if not self.manager.option(cls, 'versioning'):
                    continue

                table = self.closest_matching_table(cls)
                if table is not None:
                    builder = ModelBuilder(self.manager, cls)
                    version_cls = builder(
                        table,
                        self.manager.transaction_cls
                    )

                    self.manager.plugins.after_version_class_built(
                        cls,
                        version_cls
                    )

            self.manager.plugins.after_build_models(self.manager)

    def build_relationships(self, version_classes):
        """
        Builds relationships for all version classes.

        :param version_classes: list of generated version classes
        """
        for cls in version_classes:
            if not self.manager.option(cls, 'versioning'):
                continue

            for prop in sa.inspect(cls).iterate_properties:
                if prop.key == 'versions':
                    continue
                builder = RelationshipBuilder(self.manager, cls, prop)
                builder()

    def instrument_versioned_classes(self, mapper, cls):
        """
        Collect versioned class and add it to pending_classes list.

        :mapper mapper: SQLAlchemy mapper object
        :cls cls: SQLAlchemy declarative class
        """
        if not self.manager.options['versioning']:
            return

        if hasattr(cls, '__versioned__'):
            if (not cls.__versioned__.get('class')
                    and cls not in self.manager.pending_classes):
                self.manager.pending_classes.append(cls)
                self.manager.metadata = cls.metadata

        if hasattr(cls, '__version_parent__'):
            parent = cls.__version_parent__
            self.manager.version_class_map[parent] = cls
            self.manager.parent_class_map[cls] = parent
            del cls.__version_parent__

    def build_transaction_class(self):
        if self.manager.pending_classes:
            cls = self.manager.pending_classes[0]
            self.manager.declarative_base = get_declarative_base(cls)
            self.manager.create_transaction_model()
            self.manager.plugins.after_build_tx_class(self.manager)

    @prevent_reentry
    def configure_versioned_classes(self):
        """
        Configures all versioned classes that were collected during
        instrumentation process. The configuration has 6 steps:

        1. Build tables for version models.
        2. Build the actual version model declarative classes.
        3. Build relationships between these models.
        4. Empty pending_classes list so that consecutive mapper configuration
           does not create multiple version classes
        5. Build aliases for columns.
        6. Assign all versioned attributes to use active history.

        """
        if not self.manager.options['versioning']:
            return

        self.build_triggers()
        self.build_tables()
        self.build_transaction_class()

        if not self.manager.options['create_models']:
            self.manager.pending_classes = []
            return

        self.build_models()

        # Create copy of all pending versioned classes so that we can inspect
        # them later when creating relationships.
        pending_classes_copies = copy(self.manager.pending_classes)
        self.manager.pending_classes = []
        self.build_relationships(pending_classes_copies)
        self.enable_active_history(pending_classes_copies)
        self.create_column_aliases(pending_classes_copies)

    def enable_active_history(self, version_classes):
        """
        Assign all versioned attributes to use active history.
        """
        for cls in version_classes:
            for prop in sa.inspect(cls).iterate_properties:
                getattr(cls, prop.key).impl.active_history = True

    def create_column_aliases(self, version_classes):
        """
        Create aliases for the columns from the original model.

        This, for example, imitates the behavior of @declared_attr columns.
        """
        for cls in version_classes:
            model_mapper = sa.inspect(cls)
            version_class = self.manager.version_class_map.get(cls)
            if not version_class:
                continue

            version_class_mapper = sa.inspect(version_class)

            for key, column in model_mapper.columns.items():
                if key != column.key:
                    version_class_column = version_class.__table__.c.get(column.key)

                    if version_class_column is None:
                        continue

                    version_class_mapper.add_property(key, sa.orm.column_property(version_class_column))



"""
FlaskPlugin offers way of integrating Flask framework with
SQLAlchemy-Continuum. Flask-Plugin adds two columns for Transaction model,
namely `user_id` and `remote_addr`.

These columns are automatically populated when transaction object is created.
The `remote_addr` column is populated with the value of the remote address that
made current request. The `user_id` column is populated with the id of the
current_user object.

::

    from sqlalchemy_continuum.plugins import FlaskPlugin
    from sqlalchemy_continuum import make_versioned


    make_versioned(plugins=[FlaskPlugin()])
"""
from __future__ import absolute_import

flask = None
try:
    import flask
    from flask import request
    from flask.globals import _app_ctx_stack, _request_ctx_stack
except ImportError:
    pass
from sqlalchemy_utils import ImproperlyConfigured
from .base import Plugin


def fetch_current_user_id():
    from flask_login import current_user

    # Return None if we are outside of request context.
    if _app_ctx_stack.top is None or _request_ctx_stack.top is None:
        return
    try:
        return current_user.id
    except AttributeError:
        return


def fetch_remote_addr():
    # Return None if we are outside of request context.
    if _app_ctx_stack.top is None or _request_ctx_stack.top is None:
        return
    return request.remote_addr


class FlaskPlugin(Plugin):
    def __init__(
        self,
        current_user_id_factory=None,
        remote_addr_factory=None
    ):
        self.current_user_id_factory = (
            fetch_current_user_id if current_user_id_factory is None
            else current_user_id_factory
        )
        self.remote_addr_factory = (
            fetch_remote_addr if remote_addr_factory is None
            else remote_addr_factory
        )

        if not flask:
            raise ImproperlyConfigured(
                'Flask is required with FlaskPlugin. Please install Flask by'
                ' running pip install Flask'
            )

    def transaction_args(self, uow, session):
        return {
            'user_id': self.current_user_id_factory(),
            'remote_addr': self.remote_addr_factory()
        }



import sqlalchemy as sa

from .reverter import Reverter
from .utils import get_versioning_manager, is_internal_column, parent_class


class VersionClassBase(object):
    @property
    def previous(self):
        """
        Returns the previous version relative to this version in the version
        history. If current version is the first version this method returns
        None.
        """
        return (
            get_versioning_manager(self)
            .fetcher(parent_class(self.__class__))
            .previous(self)
        )

    @property
    def next(self):
        """
        Returns the next version relative to this version in the version
        history. If current version is the last version this method returns
        None.
        """
        return (
            get_versioning_manager(self)
            .fetcher(parent_class(self.__class__))
            .next(self)
        )

    @property
    def index(self):
        """
        Return the index of this version in the version history.
        """
        return (
            get_versioning_manager(self)
            .fetcher(parent_class(self.__class__))
            .index(self)
        )

    @property
    def changeset(self):
        """
        Return a dictionary of changed fields in this version with keys as
        field names and values as lists with first value as the old field value
        and second list value as the new value.
        """
        previous_version = self.previous
        data = {}

        for key in sa.inspect(self.__class__).columns.keys():
            if is_internal_column(self, key):
                continue
            if not previous_version:
                old = None
            else:
                old = getattr(previous_version, key)
            new = getattr(self, key)
            if old != new:
                data[key] = [old, new]

        manager = get_versioning_manager(self)
        manager.plugins.after_construct_changeset(self, data)
        return data

    def revert(self, relations=[]):
        return Reverter(self, relations=relations)()



"""
The ActivityPlugin is the most powerful plugin for tracking changes of
individual entities. If you use ActivityPlugin you probably don't need to use
TransactionChanges nor TransactionMeta plugins.

You can initalize the ActivityPlugin by adding it to versioning manager.

::

    activity_plugin = ActivityPlugin()

    make_versioned(plugins=[activity_plugin])


ActivityPlugin uses single database table for tracking activities. This table
follows the data structure in `activity stream specification`_, but it comes
with a nice twist:

    ==============  =========== =============
    Column          Type        Description
    ==============  =========== =============
    id              BigInteger  The primary key of the activity
    verb            Unicode     Verb defines the action of the activity
    data            JSON        Additional data for the activity in JSON format
    transaction_id  BigInteger  The transaction this activity was associated
                                with
    object_id       BigInteger  The primary key of the object. Object can be
                                any entity which has an integer as primary key.
    object_type     Unicode     The type of the object (class name as string)

    object_tx_id    BigInteger  The last transaction_id associated with the
                                object. This is used for efficiently fetching
                                the object version associated with this
                                activity.

    target_id       BigInteger  The primary key of the target. Target can be
                                any entity which has an integer as primary key.
    target_type     Unicode     The of the target (class name as string)

    target_tx_id    BigInteger  The last transaction_id associated with the
                                target.
    ==============  =========== =============


Each Activity has relationships to actor, object and target but it also holds
information about the associated transaction and about the last associated
transactions with the target and object. This allows each activity to also have
object_version and target_version relationships for introspecting what those
objects and targets were in given point in time. All these relationship
properties use `generic relationships`_ of the SQLAlchemy-Utils package.

Limitations
^^^^^^^^^^^

Currently all changes to parent models must be flushed or committed before
creating activities. This is due to a fact that there is still no dependency
processors for generic relationships. So when you create activities and assign
objects / targets for those please remember to flush the session before
creating an activity::


    article = Article(name=u'Some article')
    session.add(article)
    session.flush()  # <- IMPORTANT!
    first_activity = Activity(verb=u'create', object=article)
    session.add(first_activity)
    session.commit()


Targets and objects of given activity must have an integer primary key
column id.


Create activities
^^^^^^^^^^^^^^^^^


Once your models have been configured you can get the Activity model from the
ActivityPlugin class with activity_cls property::


    Activity = activity_plugin.activity_cls


Now let's say we have model called Article and Category. Each Article has one
Category. Activities should be created along with the changes you make on
these models. ::

    article = Article(name=u'Some article')
    session.add(article)
    session.flush()
    first_activity = Activity(verb=u'create', object=article)
    session.add(first_activity)
    session.commit()


Current transaction gets automatically assigned to activity object::

    first_activity.transaction  # Transaction object


Update activities
^^^^^^^^^^^^^^^^^

The object property of the Activity object holds the current object and the
object_version holds the object version at the time when the activity was
created. ::


    article.name = u'Some article updated!'
    session.flush()
    second_activity = Activity(verb=u'update', object=article)
    session.add(second_activity)
    session.commit()

    second_activity.object.name  # u'Some article updated!'
    first_activity.object.name  # u'Some article updated!'

    first_activity.object_version.name  # u'Some article'


Delete activities
^^^^^^^^^^^^^^^^^


The version properties are especially useful for delete activities. Once the
activity is fetched from the database the object is no longer available (
since its deleted), hence the only way we could show some information about the
object the user deleted is by accessing the object_version property.

::


    session.delete(article)
    session.flush()
    third_activity = Activity(verb=u'delete', object=article)
    session.add(third_activity)
    session.commit()

    third_activity.object_version.name  # u'Some article updated!'


Local version histories using targets
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The target property of the Activity model offers a way of tracking changes of
given related object. In the example below we create a new activity when adding
a category for article and then mark the article as the target of this
activity.



::


    session.add(Category(name=u'Fist category', article=article))
    session.flush()
    activity = Activity(
        verb=u'create',
        object=category,
        target=article
    )
    session.add(activity)
    session.commit()


Now if we wanted to find all the changes that affected given article we could
do so by searching through all the activities where either the object or
target is the given article.


::

    import sqlalchemy as sa


    activities = session.query(Activity).filter(
        sa.or_(
            Activity.object == article,
            Activity.target == article
        )
    )



.. _activity stream specification:
    http://www.activitystrea.ms
.. _generic relationships:
    https://sqlalchemy-utils.readthedocs.io/en/latest/generic_relationship.html
"""

import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.inspection import inspect
from sqlalchemy_utils import JSONType, generic_relationship

from .base import Plugin
from ..factory import ModelFactory
from ..utils import version_class, version_obj


class ActivityBase(object):
    id = sa.Column(
        sa.BigInteger,
        sa.schema.Sequence('activity_id_seq'),
        primary_key=True,
        autoincrement=True
    )

    verb = sa.Column(sa.Unicode(255))

    @hybrid_property
    def actor(self):
        return self.transaction.user


class ActivityFactory(ModelFactory):
    model_name = 'Activity'

    def create_class(self, manager):
        """
        Create Activity class.
        """
        class Activity(
            manager.declarative_base,
            ActivityBase
        ):
            __tablename__ = 'activity'
            manager = self

            transaction_id = sa.Column(
                sa.BigInteger,
                index=True,
                nullable=False
            )

            data = sa.Column(JSONType)

            object_type = sa.Column(sa.String(255))

            object_id = sa.Column(sa.BigInteger)

            object_tx_id = sa.Column(sa.BigInteger)

            target_type = sa.Column(sa.String(255))

            target_id = sa.Column(sa.BigInteger)

            target_tx_id = sa.Column(sa.BigInteger)

            def _calculate_tx_id(self, obj):
                session = sa.orm.object_session(self)
                if obj:
                    object_version = version_obj(session, obj)
                    if object_version:
                        return object_version.transaction_id

                    model = obj.__class__
                    version_cls = version_class(model)
                    primary_key = inspect(model).primary_key[0].name
                    return session.query(
                        sa.func.max(version_cls.transaction_id)
                    ).filter(
                        getattr(version_cls, primary_key) == getattr(obj, primary_key)
                    ).scalar()

            def calculate_object_tx_id(self):
                self.object_tx_id = self._calculate_tx_id(self.object)

            def calculate_target_tx_id(self):
                self.target_tx_id = self._calculate_tx_id(self.target)

            object = generic_relationship(
                object_type, object_id
            )

            @hybrid_property
            def object_version_type(self):
                return self.object_type + 'Version'

            @object_version_type.expression
            def object_version_type(cls):
                return sa.func.concat(cls.object_type, 'Version')

            object_version = generic_relationship(
                object_version_type, (object_id, object_tx_id)
            )

            target = generic_relationship(
                target_type, target_id
            )

            @hybrid_property
            def target_version_type(self):
                return self.target_type + 'Version'

            @target_version_type.expression
            def target_version_type(cls):
                return sa.func.concat(cls.target_type, 'Version')

            target_version = generic_relationship(
                target_version_type, (target_id, target_tx_id)
            )

        Activity.transaction = sa.orm.relationship(
            manager.transaction_cls,
            backref=sa.orm.backref(
                'activities',
            ),
            primaryjoin=(
                '%s.id == Activity.transaction_id' %
                manager.transaction_cls.__name__
            ),
            foreign_keys=[Activity.transaction_id]
        )
        return Activity


class ActivityPlugin(Plugin):
    activity_cls = None
    
    def after_build_models(self, manager):
        self.activity_cls = ActivityFactory()(manager)
        manager.activity_cls = self.activity_cls

    def is_session_modified(self, session):
        """
        Return that the session has been modified if the session contains an
        activity class.

        :param session: SQLAlchemy session object
        """
        return any(isinstance(obj, self.activity_cls) for obj in session)

    def before_flush(self, uow, session):
        for obj in session:
            if isinstance(obj, self.activity_cls):
                obj.transaction = uow.current_transaction
                obj.calculate_target_tx_id()
                obj.calculate_object_tx_id()

    def after_version_class_built(self, parent_cls, version_cls):
        pass



"""
The PropertyModTrackerPlugin offers a way of efficiently tracking individual
property modifications. With PropertyModTrackerPlugin you can make efficient
queries such as:

Find all versions of model X where user updated the property A or property B.

Find all versions of model X where user didn't update property A.

PropertyModTrackerPlugin adds separate modified tracking column for each
versioned column. So for example if you have versioned model called Article
with columns `name` and `content`, this plugin would add two additional boolean
columns `name_mod` and `content_mod` for the version model. When user commits
transactions the plugin automatically updates these boolean columns.
"""

from copy import copy
import sqlalchemy as sa
from sqlalchemy_utils.functions import has_changes
from .base import Plugin
from ..utils import versioned_column_properties


class PropertyModTrackerPlugin(Plugin):
    column_suffix = '_mod'

    def create_mod_column(self, column):
        return sa.Column(
            column.name + self.column_suffix,
            sa.Boolean,
            key=column.key + self.column_suffix,
            default=False,
            server_default=sa.sql.expression.false(),
            nullable=False
        )

    def after_build_version_table_columns(self, table_builder, columns):
        # Only create modification tracking columns for tables that are
        # associated with actual model classes. In other words do not create
        # mod tracking columns for association tables.
        if table_builder.model:
            for column in table_builder.parent_table.c:
                if not table_builder.manager.is_excluded_column(
                    table_builder.model, column
                ) and not column.primary_key:
                    columns.append(self.create_mod_column(column))

    def after_create_version_object(self, uow, parent_obj, version_obj):
        session = sa.orm.object_session(parent_obj)
        is_deleted = parent_obj in session.deleted

        for prop in versioned_column_properties(parent_obj):
            if has_changes(parent_obj, prop.key) or is_deleted:
                setattr(
                    version_obj,
                    prop.key + self.column_suffix,
                    True
                )

    def after_construct_changeset(self, version_obj, changeset):
        for key in copy(changeset).keys():
            if key.endswith(self.column_suffix):
                del changeset[key]



from .base import Plugin
from ..operation import Operation
from ..utils import versioned_column_properties, is_internal_column


class NullDeletePlugin(Plugin):
    def should_nullify_column(self, version_obj, prop):
        """
        Return whether or not given column of given version object should
        be nullified (set to None) at the end of the transaction.

        :param version_obj:
            Version object to check the attribute nullification
        :paremt prop:
            SQLAlchemy ColumnProperty object
        """
        return (
            version_obj.operation_type == Operation.DELETE and
            not prop.columns[0].primary_key and
            not is_internal_column(version_obj, prop.key)
        )

    def after_create_version_object(self, uow, parent_obj, version_obj):
        for prop in versioned_column_properties(parent_obj):
            if self.should_nullify_column(version_obj, prop):
                setattr(version_obj, prop.key, None)



import sqlalchemy as sa
from .operation import Operation
from .utils import versioned_column_properties, parent_class


def first_level(paths):
    for path in paths:
        yield path.split('.')[0]


def subpaths(paths, name):
    for path in paths:
        parts = path.split('.')
        if len(parts) > 1 and parts[0] == name:
            yield '.'.join(parts[1:])


class ReverterException(Exception):
    pass


class Reverter(object):
    def __init__(self, obj, visited_objects=None, relations=[]):
        self.visited_objects = visited_objects or []
        self.obj = obj
        self.version_parent = self.obj.version_parent
        self.parent_class = parent_class(self.obj.__class__)
        self.parent_mapper = sa.inspect(self.parent_class)
        self.session = sa.orm.object_session(self.obj)

        self.relations = list(relations)
        for path in relations:
            subpath = path.split('.')[0]
            if subpath not in self.parent_mapper.relationships:
                raise ReverterException(
                    "Could not initialize Reverter. Class '%s' does not have "
                    "relationship '%s'." % (
                        parent_class(self.obj.__class__).__name__,
                        subpath
                    )
                )

    def revert_properties(self):
        for prop in versioned_column_properties(self.parent_class):
            setattr(
                self.version_parent,
                prop.key,
                getattr(self.obj, prop.key)
            )

    def revert_association(self, prop):
        if prop.uselist:
            setattr(self.version_parent, prop.key, [])
            for child_obj in getattr(self.obj, prop.key):
                value = self.revert_child(child_obj, prop)
                if value:
                    getattr(self.version_parent, prop.key).append(
                        value
                    )
        else:
            setattr(self.version_parent, prop.key, None)
            value = getattr(self.obj, prop.key)
            value = self.revert_child(
                value, prop
            )
            if value:
                setattr(self.version_parent, prop.key, value)

    def revert_relationship(self, prop):
        if prop.secondary is not None:
            self.revert_association(prop)
        else:
            if prop.uselist:
                values = []
                for child_obj in getattr(self.obj, prop.key):
                    value = self.revert_child(child_obj, prop)
                    if value:
                        values.append(value)

                for value in getattr(self.version_parent, prop.key, []):
                    if value not in values:
                        self.session.delete(value)
            else:
                self.revert_child(getattr(self.obj, prop.key), prop)

    def revert_child(self, child, prop):
        return self.__class__(
            child,
            visited_objects=self.visited_objects,
            relations=subpaths(self.relations, prop.key)
        )()

    def revert_relationships(self):
        for prop in self.parent_mapper.iterate_properties:
            if isinstance(prop, sa.orm.RelationshipProperty):
                if prop.key in ['versions', 'transaction']:
                    continue

                if prop.key not in first_level(self.relations):
                    continue

                self.revert_relationship(prop)

    def __call__(self):
        if self.obj in self.visited_objects:
            return (
                None if self.obj.operation_type == Operation.DELETE
                else self.version_parent
            )

        if self.obj.operation_type == Operation.DELETE:
            self.session.delete(self.version_parent)
            return

        self.visited_objects.append(self.obj)

        # Check if parent object has been deleted
        if self.version_parent is None:
            self.version_parent = parent_class(self.obj.__class__)()

        # Before reifying relations we need to reify object properties. This
        # is needed because reifying relations might need to flush the session
        # which leads to errors when sqlalchemy tries to insert null values
        # into parent object (if parent object has not null constraints).
        self.revert_properties()
        self.revert_relationships()
        self.session.add(self.version_parent)

        return self.version_parent



import re
from functools import wraps

import sqlalchemy as sa
from sqlalchemy.orm import object_session
from sqlalchemy_utils import get_column_key

from .builder import Builder
from .fetcher import SubqueryFetcher, ValidityFetcher
from .operation import Operation
from .plugins import PluginCollection
from .transaction import TransactionFactory
from .unit_of_work import UnitOfWork
from .utils import is_modified, is_versioned


def tracked_operation(func):
    @wraps(func)
    def wrapper(self, mapper, connection, target):
        if not is_versioned(target):
            return
        session = object_session(target)
        conn = session.connection()
        try:
            uow = self.units_of_work[conn]
        except KeyError:
            try:
                uow = self.units_of_work[conn.engine]
            except KeyError:
                for connection in self.units_of_work.keys():
                    if not connection.closed and connection.connection is conn.connection:
                        uow = self.unit_of_work(session)
                        break  # The ConnectionFairy is the same, this connection is a clone
                else:
                    raise
        return func(self, uow, target)
    return wrapper


class VersioningManager(object):
    """
    VersioningManager delegates versioning configuration operations to builder
    classes and the actual versioning to UnitOfWork class. Manager contains
    configuration options that act as defaults for all versioned classes.

    :param unit_of_work_cls:
        The UnitOfWork class to use for initializing UnitOfWork objects for
        versioning
    :param transaction_cls:
        Transaction class to use for versioning. If None, the default
        Transaction class generated by TransactionFactory will be used.
    :param user_cls:
        User class which Transaction class should have relationship to. This
        can either be a class or string name of a class for lazy evaluation.
    :param options:
        Versioning options
    :param plugins:
        Versioning plugins that listen the events invoked by the manager.
    :param builder:
        Builder object which handles the building of versioning tables and
        models.
    """
    def __init__(
        self,
        unit_of_work_cls=UnitOfWork,
        transaction_cls=None,
        user_cls=None,
        options={},
        plugins=None,
        builder=None
    ):
        self.uow_class = unit_of_work_cls
        if builder is None:
            self.builder = Builder()
        else:
            self.builder = builder
        self.builder.manager = self
        self.reset()
        if transaction_cls is not None:
            self.transaction_cls = transaction_cls
        else:
            self.transaction_cls = TransactionFactory()
        if user_cls is not None:
            self.user_cls = user_cls

        self.options = {
            'versioning': True,
            'base_classes': None,
            'table_name': '%s_version',
            'exclude': [],
            'include': [],
            'native_versioning': False,
            'create_models': True,
            'create_tables': True,
            'transaction_column_name': 'transaction_id',
            'end_transaction_column_name': 'end_transaction_id',
            'operation_type_column_name': 'operation_type',
            'strategy': 'validity',
            'use_module_name': False
        }
        if plugins is None:
            self.plugins = []
        else:
            self.plugins = plugins
        self.options.update(options)

    @property
    def plugins(self):
        return self._plugins

    @plugins.setter
    def plugins(self, plugin_collection):
        self._plugins = PluginCollection(plugin_collection)

    def fetcher(self, obj):
        if self.option(obj, 'strategy') == 'subquery':
            return SubqueryFetcher(self)
        else:
            return ValidityFetcher(self)

    def reset(self):
        """
        Resets this manager's internal state.

        This method should be used in test cases that create models on the fly.
        Otherwise history_class_map and some other variables would be polluted
        by no more used model classes.
        """
        self.tables = {}
        self.pending_classes = []
        self.association_tables = set()
        self.association_version_tables = set()
        self.declarative_base = None
        self.version_class_map = {}
        self.parent_class_map = {}
        self.session_listeners = {
            'before_flush': self.before_flush,
            'after_flush': self.after_flush,
            'after_commit': self.clear,
            'after_rollback': self.clear,
        }
        self.mapper_listeners = {
            'after_delete': self.track_deletes,
            'after_update': self.track_updates,
            'after_insert': self.track_inserts,
        }
        self.class_config_listeners = {
            'instrument_class': self.builder.instrument_versioned_classes,
            'after_configured': self.builder.configure_versioned_classes,
        }

        # A dictionary of units of work. Keys as connection objects and values
        # as UnitOfWork objects.
        self.units_of_work = {}

        self.session_connection_map = {}

        self.metadata = None

    def create_transaction_model(self):
        """
        Create Transaction class but only if it doesn't already exist in
        declarative model registry.
        """
        if isinstance(self.transaction_cls, TransactionFactory):
            self.transaction_cls = self.transaction_cls(self)
        return self.transaction_cls

    def is_excluded_column(self, model, column):
        try:
            key = get_column_key(model, column)
        except sa.orm.exc.UnmappedColumnError:
            return False

        return self.is_excluded_property(model, key)

    def is_excluded_property(self, model, key):
        """
        Returns whether or not given property of given model is excluded from
        the associated history model.

        :param model: SQLAlchemy declarative model object.
        :param key: Model property key
        """
        if key in self.option(model, 'include'):
            return False
        return key in self.option(model, 'exclude')

    def option(self, model, name):
        """
        Returns the option value for given model. If the option is not found
        from given model falls back to default values of this manager object.
        If the option is not found from this manager object either this method
        throws a KeyError.

        :param model: SQLAlchemy declarative object
        :param name: name of the versioning option
        """
        if not hasattr(model, '__versioned__'):
            raise TypeError('Model %r is not versioned.' % model)
        try:
            return model.__versioned__[name]
        except KeyError:
            return self.options[name]

    def apply_class_configuration_listeners(self, mapper):
        """
        Applies class configuration listeners for given mapper.

        The listener work in two phases:

        1. Class instrumentation phase
            The first listeners listens to class instrumentation event and
            handles the collecting of versioned models and adds them to
            the pending_classes list.
        2. After class configuration phase
            The second listener listens to after class configuration event and
            handles the actual history model generation based on list that
            was collected during class instrumenation phase.

        :param mapper:
            SQLAlchemy mapper to apply the class configuration listeners to
        """
        for event_name, listener in self.class_config_listeners.items():
            sa.event.listen(mapper, event_name, listener)

    def remove_class_configuration_listeners(self, mapper):
        """
        Remove versioning class configuration listeners from specified mapper.

        :param mapper:
            mapper to remove class configuration listeners from
        """
        for event_name, listener in self.class_config_listeners.items():
            sa.event.remove(mapper, event_name, listener)

    def track_operations(self, mapper):
        """
        Attach listeners for specified mapper that track SQL inserts, updates
        and deletes.

        :param mapper: mapper to track the SQL operations from
        """
        for event_name, listener in self.mapper_listeners.items():
            sa.event.listen(mapper, event_name, listener)

    def remove_operations_tracking(self, mapper):
        """
        Remove listeners from specified mapper that track SQL inserts, updates
        and deletes.

        :param mapper:
            mapper to remove the SQL operations tracking listeners from
        """
        for event_name, listener in self.mapper_listeners.items():
            sa.event.remove(mapper, event_name, listener)

    def track_session(self, session):
        """
        Attach listeners that track the operations (flushing, committing and
        rolling back) of given session. This method should be used in
        conjunction with `track_operations`.

        :param session: SQLAlchemy session to track the operations from
        """
        for event_name, listener in self.session_listeners.items():
            sa.event.listen(session, event_name, listener)

    def remove_session_tracking(self, session):
        """
        Remove listeners that track the operations (flushing, committing and
        rolling back) of given session. This method should be used in
        conjunction with `remove_operations_tracking`.

        :param session:
            SQLAlchemy session to remove the operations tracking from
        """
        for event_name, listener in self.session_listeners.items():
            sa.event.remove(session, event_name, listener)

    @tracked_operation
    def track_inserts(self, uow, target):
        """
        Track object insert operations. Whenever object is inserted it is
        added to this UnitOfWork's internal operations dictionary.
        """
        uow.operations.add_insert(target)

    @tracked_operation
    def track_updates(self, uow, target):
        """
        Track object update operations. Whenever object is updated it is
        added to this UnitOfWork's internal operations dictionary.
        """
        if not is_modified(target):
            return
        uow.operations.add_update(target)

    @tracked_operation
    def track_deletes(self, uow, target):
        """
        Track object deletion operations. Whenever object is deleted it is
        added to this UnitOfWork's internal operations dictionary.
        """
        uow.operations.add_delete(target)

    def unit_of_work(self, session):
        """
        Return the associated SQLAlchemy-Continuum UnitOfWork object for given
        SQLAlchemy session object.

        If no UnitOfWork object exists for given object then this method tries
        to create one.

        :param session: SQLAlchemy session object
        """
        conn = session.connection()
        if conn not in self.session_connection_map.values():
            self.session_connection_map[session] = conn

        if conn in self.units_of_work:
            return self.units_of_work[conn]
        else:
            uow = self.uow_class(self)
            self.units_of_work[conn] = uow
            return uow

    def before_flush(self, session, flush_context, instances):
        """
        Before flush listener for SQLAlchemy sessions. If this manager has
        versioning enabled this listener invokes the process before flush of
        associated UnitOfWork object.

        :param session: SQLAlchemy session
        """
        if not self.options['versioning']:
            return

        uow = self.unit_of_work(session)
        uow.process_before_flush(session)

    def after_flush(self, session, flush_context):
        """
        After flush listener for SQLAlchemy sessions. If this manager has
        versioning enabled this listener gets the UnitOfWork associated with
        session's connections and invokes the process_after_flush method
        of that object.

        :param session: SQLAlchemy session
        """
        if not self.options['versioning']:
            return
        uow = self.unit_of_work(session)
        uow.process_after_flush(session)

    def clear(self, session):
        """
        Simple SQLAlchemy listener that is being invoked after successful
        transaction commit or when transaction rollback occurs. The purpose of
        this listener is to reset this UnitOfWork back to its initialization
        state.

        :param session: SQLAlchemy session object
        """
        if session.transaction.nested:
            return
        conn = self.session_connection_map.pop(session, None)
        if conn is None:
            return

        if conn in self.units_of_work:
            uow = self.units_of_work[conn]
            uow.reset(session)
            del self.units_of_work[conn]

        for connection in dict(self.units_of_work).keys():
            if connection.closed or conn.connection is connection.connection:
                uow = self.units_of_work[connection]
                uow.reset(session)
                del self.units_of_work[connection]

    def clear_connection(self, conn):
        if conn in self.units_of_work:
            uow = self.units_of_work[conn]
            uow.reset()
            del self.units_of_work[conn]


        for session, connection in dict(self.session_connection_map).items():
            if connection is conn:
                del self.session_connection_map[session]


        for connection in dict(self.units_of_work).keys():
            if connection.closed or conn.connection is connection.connection:
                uow = self.units_of_work[connection]
                uow.reset()
                del self.units_of_work[connection]


    def append_association_operation(self, conn, table_name, params, op):
        """
        Append history association operation to pending_statements list.
        """
        stmt = (
            self.metadata.tables[self.options['table_name'] % table_name]
            .insert()
            .values({**params, 'operation_type': op})
        )
        try:
            uow = self.units_of_work[conn]
        except KeyError:
            try:
                uow = self.units_of_work[conn.engine]
            except KeyError:
                for connection in self.units_of_work.keys():
                    if not connection.closed and connection.connection is conn.connection:
                        uow = self.unit_of_work(conn.session)
                        break  # The ConnectionFairy is the same, this connection is a clone
                else:
                    raise
        uow.pending_statements.append(stmt)

    def track_cloned_connections(self, c, opt):
        """
        Track cloned connections from association tables.
        """
        if c not in self.units_of_work.keys():
            for connection, uow in dict(self.units_of_work).items():
                if not connection.closed and connection.connection is c.connection:  # ConnectionFairy is the same - this is a clone
                    self.units_of_work[c] = uow

    def track_association_operations(
        self, conn, cursor, statement, parameters, context, executemany
    ):
        """
        Track association operations and adds the generated history
        association operations to pending_statements list.
        """
        if (
            not self.options['versioning'] and
            not self.options['native_versioning']
        ):
            return

        op = None
        if context.isinsert:
            op = Operation.INSERT
        elif context.isdelete:
            op = Operation.DELETE

        if op is not None:
            table_name = statement.split(' ')[2]
            table_names = [
                table.name if not table.schema else table.schema + '.' + table.name
                for table in self.association_tables
            ]
            if table_name in table_names:
                if executemany:
                    # SQLAlchemy does not support function based values for
                    # multi-inserts, hence we need to convert the orignal
                    # multi-insert into batch of normal inserts
                    for params in parameters:
                        self.append_association_operation(
                            conn,
                            table_name,
                            self.positional_args_to_dict(
                                op, statement, params
                            ),
                            op
                        )
                else:
                    self.append_association_operation(
                        conn,
                        table_name,
                        self.positional_args_to_dict(
                            op,
                            statement,
                            parameters
                        ),
                        op
                    )

    def positional_args_to_dict(self, op, statement, params):
        """
        On some drivers (eg sqlite) generated INSERT statements use positional
        args instead of key value dictionary. This method converts positional
        args to key value dict.

        :param statement: SQL statement string
        :param params: tuple or dict of statement parameters
        """
        if isinstance(params, tuple):
            parameters = {}
            if op == Operation.DELETE:
                regexp = '^DELETE FROM (.+?) WHERE'
                match = re.match(regexp, statement)
                tablename = match.groups()[0].strip('"').strip("'").strip('`')
                table = self.metadata.tables[tablename]
                columns = table.primary_key.columns.values()
                for index, column in enumerate(columns):
                    parameters[column.name] = params[index]
            else:
                columns = [
                    column.strip() for column in
                    statement.split('(')[1].split(')')[0].split(',')
                ]
                for index, column in enumerate(columns):
                    parameters[column] = params[index]
            return parameters
        return params



from itertools import chain
from inspect import isclass
from collections import defaultdict

import sqlalchemy as sa
from sqlalchemy.orm.attributes import get_history
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy_utils.functions import (
    get_primary_keys,
    identity,
    naturally_equivalent,
)

from .exc import ClassNotVersioned


def get_versioning_manager(obj_or_class):
    """
    Return the associated SQLAlchemy-Continuum VersioningManager for given
    SQLAlchemy declarative model class or object.

    :param obj_or_class: SQLAlchemy declarative model object or class
    """
    if isinstance(obj_or_class, AliasedClass):
        obj_or_class = sa.inspect(obj_or_class).mapper.class_
    cls = obj_or_class if isclass(obj_or_class) else obj_or_class.__class__
    try:
        return cls.__versioning_manager__
    except AttributeError:
        raise ClassNotVersioned(cls.__name__)


def option(obj_or_class, option_name):
    """
    Return the option value of given option for given versioned object or
    class.

    :param obj_or_class: SQLAlchemy declarative model object or class
    :param option_name: The name of an option to return
    """
    if isinstance(obj_or_class, AliasedClass):
        obj_or_class = sa.inspect(obj_or_class).mapper.class_
    cls = obj_or_class if isclass(obj_or_class) else obj_or_class.__class__
    if not hasattr(cls, '__versioned__'):
        cls = parent_class(cls)
    return get_versioning_manager(cls).option(
        cls, option_name
    )


def tx_column_name(obj):
    return option(obj, 'transaction_column_name')


def end_tx_column_name(obj):
    return option(obj, 'end_transaction_column_name')


def end_tx_attr(obj):
    return getattr(
        obj.__class__,
        end_tx_column_name(obj)
    )


def parent_class(version_cls):
    """
    Return the parent class for given version model class.

    ::

        parent_class(ArticleVersion)  # Article class


    :param model: SQLAlchemy declarative version model class

    .. seealso:: :func:`version_class`
    """
    return get_versioning_manager(version_cls).parent_class_map[version_cls]


def transaction_class(cls):
    """
    Return the associated transaction class for given versioned SQLAlchemy
    declarative class or version class.

    ::

        from sqlalchemy_continuum import transaction_class


        transaction_class(Article)  # Transaction class

    :param cls: SQLAlchemy versioned declarative class or version model class
    """
    return get_versioning_manager(cls).transaction_cls


def version_obj(session, parent_obj):
    manager = get_versioning_manager(parent_obj)
    uow = manager.unit_of_work(session)
    for version_obj in uow.version_session:
        if (
            parent_class(version_obj.__class__) == parent_obj.__class__ and
            identity(version_obj)[:-1] == identity(parent_obj)
        ):
            return version_obj


def version_class(model):
    """
    Return the version class for given SQLAlchemy declarative model class.

    ::

        version_class(Article)  # ArticleVersion class


    :param model: SQLAlchemy declarative model class

    .. seealso:: :func:`parent_class`
    """
    manager = get_versioning_manager(model)
    try:
        return manager.version_class_map[model]
    except KeyError:
        return model


def version_table(table):
    """
    Return associated version table for given SQLAlchemy Table object.

    :param table: SQLAlchemy Table object
    """
    if table.schema:
        return table.metadata.tables[
            table.schema + '.' + table.name + '_version'
        ]
    elif table.metadata.schema:
        return table.metadata.tables[
            table.metadata.schema + '.' + table.name + '_version'
        ]
    else:
        return table.metadata.tables[
            table.name + '_version'
        ]


def versioned_objects(session):
    """
    Return all versioned objects in given session.

    :param session: SQLAlchemy session object

    .. seealso:: :func:`is_versioned`
    """
    for obj in session:
        if is_versioned(obj):
            yield obj


def is_versioned(obj_or_class):
    """
    Return whether or not given object is versioned.

    ::

        is_versioned(Article)  # True

        article = Article()

        is_versioned(article)  # True


    :param obj_or_class:
        SQLAlchemy declarative model object or SQLAlchemy declarative model
        class.

    .. seealso:: :func:`versioned_objects`
    """
    try:
        return (
            hasattr(obj_or_class, '__versioned__') and
            get_versioning_manager(obj_or_class).option(
                obj_or_class, 'versioning'
            )
        )
    except ClassNotVersioned:
        return False


def versioned_column_properties(obj_or_class):
    """
    Return all versioned column properties for given versioned SQLAlchemy
    declarative model object.

    :param obj: SQLAlchemy declarative model object
    """
    manager = get_versioning_manager(obj_or_class)

    cls = obj_or_class if isclass(obj_or_class) else obj_or_class.__class__

    mapper = sa.inspect(cls)
    for key, column in mapper.columns.items():
        # Ignores non table columns
        if not is_table_column(column):
            continue

        if not manager.is_excluded_property(obj_or_class, key):
            yield getattr(mapper.attrs, key)


def versioned_relationships(obj, versioned_column_keys):
    """
    Return all versioned relationships for given versioned SQLAlchemy
    declarative model object.

    :param obj: SQLAlchemy declarative model object
    """
    for prop in sa.inspect(obj.__class__).relationships:
        if any(c.key in versioned_column_keys for c in prop.local_columns):
            yield prop


def vacuum(session, model, yield_per=1000):
    """
    When making structural changes to version tables (for example dropping
    columns) there are sometimes situations where some old version records
    become futile.

    Vacuum deletes all futile version rows which had no changes compared to
    previous version.


    ::


        from sqlalchemy_continuum import vacuum


        vacuum(session, User)  # vacuums user version


    :param session: SQLAlchemy session object
    :param model: SQLAlchemy declarative model class
    :param yield_per: how many rows to process at a time
    """
    version_cls = version_class(model)
    versions = defaultdict(list)

    query = (
        session.query(version_cls)
        .order_by(option(version_cls, 'transaction_column_name'))
    ).yield_per(yield_per)

    primary_key_col = sa.inspection.inspect(model).primary_key[0].name

    for version in query:
        version_id = getattr(version, primary_key_col)
        if versions[version_id]:
            prev_version = versions[version_id][-1]
            if naturally_equivalent(prev_version, version):
                session.delete(version)
        else:
            versions[version_id].append(version)


def is_table_column(column):
    """
    Return wheter of not give field is a column over the database table.

    :param column: SQLAclhemy model field.
    :rtype: bool
    """
    return isinstance(column, sa.Column)


def is_internal_column(model, column_name):
    """
    Return whether or not given column of given SQLAlchemy declarative classs
    is considered an internal column (a column whose purpose is mainly
    for SA-Continuum's internal use).

    :param version_obj: SQLAlchemy declarative class
    :param column_name: Name of the column
    """
    return column_name in (
        option(model, 'transaction_column_name'),
        option(model, 'end_transaction_column_name'),
        option(model, 'operation_type_column_name')
    )


def is_modified_or_deleted(obj):
    """
    Return whether or not some of the versioned properties of given SQLAlchemy
    declarative object have been modified or if the object has been deleted.

    :param obj: SQLAlchemy declarative model object
    """
    session = sa.orm.object_session(obj)
    return is_versioned(obj) and (
        is_modified(obj) or
        obj in chain(session.deleted, session.new)
    )


def is_modified(obj):
    """
    Return whether or not the versioned properties of given object have been
    modified.

    ::

        article = Article()

        is_modified(article)  # False

        article.name = 'Something'

        is_modified(article)  # True


    :param obj: SQLAlchemy declarative model object

    .. seealso:: :func:`is_modified_or_deleted`
    .. seealso:: :func:`is_session_modified`
    """
    column_names = sa.inspect(obj.__class__).columns.keys()
    versioned_column_keys = [
        prop.key for prop in versioned_column_properties(obj)
    ]
    versioned_relationship_keys = [
        prop.key
        for prop in versioned_relationships(obj, versioned_column_keys)
    ]
    for key, attr in sa.inspect(obj).attrs.items():
        if key in column_names:
            if key not in versioned_column_keys:
                continue
            if attr.history.has_changes():
                return True
        if key in versioned_relationship_keys:
            if attr.history.has_changes():
                return True
    return False


def is_session_modified(session):
    """
    Return whether or not any of the versioned objects in given session have
    been either modified or deleted.

    :param session: SQLAlchemy session object

    .. seealso:: :func:`is_versioned`
    .. seealso:: :func:`versioned_objects`
    """
    return any(
        is_modified_or_deleted(obj) for obj in versioned_objects(session)
    )


def count_versions(obj):
    """
    Return the number of versions given object has. This function works even
    when obj has `create_models` and `create_tables` versioned settings
    disabled.

    ::

        article = Article(name=u'Some article')

        count_versions(article)  # 0

        session.add(article)
        session.commit()

        count_versions(article)  # 1


    :param obj: SQLAlchemy declarative model object
    """
    session = sa.orm.object_session(obj)
    if session is None:
        # If object is transient, we assume it has no version history.
        return 0
    manager = get_versioning_manager(obj)
    table_name = manager.option(obj, 'table_name') % obj.__table__.name
    criteria = [
        '%s = %r' % (pk, getattr(obj, pk))
        for pk in get_primary_keys(obj)
    ]
    query = 'SELECT COUNT(1) FROM %s WHERE %s' % (
        table_name,
        ' AND '.join(criteria)
    )
    return session.execute(query).scalar()


def changeset(obj):
    """
    Return a humanized changeset for given SQLAlchemy declarative object. With
    this function you can easily check the changeset of given object in current
    transaction.

    ::


        from sqlalchemy_continuum import changeset


        article = Article(name=u'Some article')
        changeset(article)
        # {'name': [u'Some article', None]}

    :param obj: SQLAlchemy declarative model object
    """
    data = {}
    session = sa.orm.object_session(obj)
    if session and obj in session.deleted:
        columns = [c for c in sa.inspect(obj.__class__).columns.values()
                   if is_table_column(c)]

        for column in columns:
            if not column.primary_key:
                value = getattr(obj, column.key)
                if value is not None:
                    data[column.key] = [None, getattr(obj, column.key)]
    else:
        for prop in obj.__mapper__.iterate_properties:
            history = get_history(obj, prop.key)
            if history.has_changes():
                old_value = history.deleted[0] if history.deleted else None
                new_value = history.added[0] if history.added else None

                if new_value:
                    data[prop.key] = [new_value, old_value]
    return data



class VersioningClauseAdapter(sa.sql.visitors.ReplacingCloningVisitor):
    def replace(self, col):
        if isinstance(col, sa.Column):
            table = version_table(col.table)
            return table.c.get(col.key)


def adapt_columns(expr):
    return VersioningClauseAdapter().traverse(expr)



from .activity import ActivityPlugin
from .base import Plugin, PluginCollection
from .flask import FlaskPlugin
from .null_delete import NullDeletePlugin
from .property_mod_tracker import PropertyModTrackerPlugin
from .transaction_changes import TransactionChangesPlugin
from .transaction_meta import TransactionMetaPlugin



import sqlalchemy as sa
from .exc import ClassNotVersioned, ImproperlyConfigured
from .manager import VersioningManager
from .operation import Operation
from .transaction import TransactionFactory
from .unit_of_work import UnitOfWork
from .utils import (
    changeset,
    count_versions,
    get_versioning_manager,
    is_modified,
    is_session_modified,
    parent_class,
    transaction_class,
    tx_column_name,
    vacuum,
    version_class,
)


__version__ = '1.3.14'


versioning_manager = VersioningManager()


def make_versioned(
    mapper=sa.orm.mapper,
    session=sa.orm.session.Session,
    manager=versioning_manager,
    plugins=None,
    options=None,
    user_cls='User'
):
    """
    This is the public API function of SQLAlchemy-Continuum for making certain
    mappers and sessions versioned. By default this applies to all mappers and
    all sessions.

    :param mapper:
        SQLAlchemy mapper to apply the versioning to.
    :param session:
        SQLAlchemy session to apply the versioning to. By default this is
        sa.orm.session.Session meaning it applies to all Session subclasses.
    :param manager:
        SQLAlchemy-Continuum versioning manager.
    :param plugins:
        Plugins to pass for versioning manager.
    :param options:
        A dictionary of VersioningManager options.
    :param user_cls:
        User class which the Transaction class should have relationship to.
        This can either be a class or string name of a class for lazy
        evaluation.
    """
    if plugins is not None:
        manager.plugins = plugins

    if options is not None:
        manager.options.update(options)

    manager.user_cls = user_cls
    manager.apply_class_configuration_listeners(mapper)
    manager.track_operations(mapper)
    manager.track_session(session)

    sa.event.listen(
        sa.engine.Engine,
        'before_cursor_execute',
        manager.track_association_operations
    )

    sa.event.listen(
        sa.engine.Engine,
        'rollback',
        manager.clear_connection
    )

    sa.event.listen(
        sa.engine.Engine,
        'set_connection_execution_options',
        manager.track_cloned_connections
    )


def remove_versioning(
    mapper=sa.orm.mapper,
    session=sa.orm.session.Session,
    manager=versioning_manager
):
    """
    Remove the versioning from given mapper / session and manager.

    :param mapper:
        SQLAlchemy mapper to remove the versioning from.
    :param session:
        SQLAlchemy session to remove the versioning from. By default this is
        sa.orm.session.Session meaning it applies to all sessions.
    :param manager:
        SQLAlchemy-Continuum versioning manager.
    """
    manager.reset()
    manager.remove_class_configuration_listeners(mapper)
    manager.remove_operations_tracking(mapper)
    manager.remove_session_tracking(session)
    sa.event.remove(
        sa.engine.Engine,
        'before_cursor_execute',
        manager.track_association_operations
    )

    sa.event.remove(
        sa.engine.Engine,
        'rollback',
        manager.clear_connection
    )

    sa.event.remove(
        sa.engine.Engine,
        'set_connection_execution_options',
        manager.track_cloned_connections
    )



"""
TransactionChanges provides way of keeping track efficiently which declarative
models were changed in given transaction. This can be useful when transactions
need to be queried afterwards for problems such as:

1. Find all transactions which affected `User` model.

2. Find all transactions which didn't affect models `Entity` and `Event`.

The plugin works in two ways. On class instrumentation phase this plugin
creates a special transaction model called `TransactionChanges`. This model is
associated with table called `transaction_changes`, which has only only two
fields: transaction_id and entity_name. If for example transaction consisted
of saving 5 new User entities and 1 Article entity, two new rows would be
inserted into transaction_changes table.

================    =================
transaction_id          entity_name
----------------    -----------------
233678                  User
233678                  Article
================    =================
"""
import six
import sqlalchemy as sa

from .base import Plugin
from ..factory import ModelFactory
from ..utils import option


class TransactionChangesBase(object):
    transaction_id = sa.Column(
        sa.BigInteger,
        primary_key=True
    )
    entity_name = sa.Column(sa.Unicode(255), primary_key=True)


class TransactionChangesFactory(ModelFactory):
    model_name = 'TransactionChanges'

    def create_class(self, manager):
        """
        Create TransactionChanges class.
        """
        class TransactionChanges(
            manager.declarative_base,
            TransactionChangesBase
        ):
            __tablename__ = 'transaction_changes'

        TransactionChanges.transaction = sa.orm.relationship(
            manager.transaction_cls,
            backref=sa.orm.backref(
                'changes',
            ),
            primaryjoin=(
                '%s.id == TransactionChanges.transaction_id' %
                manager.transaction_cls.__name__
            ),
            foreign_keys=[TransactionChanges.transaction_id]
        )
        return TransactionChanges


class TransactionChangesPlugin(Plugin):
    objects = None

    def after_build_tx_class(self, manager):
        self.model_class = TransactionChangesFactory()(manager)

    def after_build_models(self, manager):
        self.model_class = TransactionChangesFactory()(manager)

    def before_create_version_objects(self, uow, session):
        for entity in uow.operations.entities:
            params = uow.current_transaction.id, six.text_type(entity.__name__)
            changes = session.query(self.model_class).get(params)
            if not changes:
                changes = self.model_class(
                    transaction_id=uow.current_transaction.id,
                    entity_name=six.text_type(entity.__name__)
                )
                session.add(changes)

    def clear(self):
        self.objects = None

    def after_rollback(self, uow, session):
        self.clear()

    def ater_commit(self, uow, session):
        self.clear()

    def after_version_class_built(self, parent_cls, version_cls):
        parent_cls.__versioned__['transaction_changes'] = self.model_class



from copy import copy
import six
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import column_property
from sqlalchemy_utils.functions import get_declarative_base

from .utils import adapt_columns, option
from .version import VersionClassBase


def find_closest_versioned_parent(manager, model):
    """
    Finds the closest versioned parent for current parent model.
    """
    for class_ in model.__bases__:
        if class_ in manager.version_class_map:
            return manager.version_class_map[class_]

def versioned_parents(manager, model):
    """
    Finds all versioned ancestors for current parent model.
    """
    for class_ in model.__mro__:
        if class_ in manager.version_class_map:
            yield manager.version_class_map[class_]


def get_base_class(manager, model):
    """
    Returns all base classes for history model.
    """
    return (
        option(model, 'base_classes')
        or
        (get_declarative_base(model), )
    )


def version_base(manager, parent_cls, base_class_factory=None):
    if base_class_factory is None:
        base_class_factory = get_base_class

    VersionBase = find_closest_versioned_parent(manager, parent_cls)

    if not VersionBase:
        VersionBase = type(
            'VersionBase',
            (base_class_factory(manager, parent_cls) + (VersionClassBase, )),
            {'__abstract__': True}
        )

    return VersionBase


def copy_mapper_args(model):
    args = {}
    if hasattr(model, '__mapper_args__'):
        arg_names = (
            'with_polymorphic',
            'polymorphic_identity',
            'concrete'
        )
        for arg in arg_names:
            if arg in model.__mapper_args__:
                args[arg] = (
                    model.__mapper_args__[arg]
                )

        if 'order_by' in model.__mapper_args__:
            arg = model.__mapper_args__['order_by']
            # Only allow string based order_by reflection to version
            # classes.
            if isinstance(arg, six.string_types):
                args['order_by'] = arg

        if 'polymorphic_on' in model.__mapper_args__:
            column = model.__mapper_args__['polymorphic_on']
            if isinstance(column, six.string_types):
                args['polymorphic_on'] = column
            else:
                args['polymorphic_on'] = column.key
    return args


class ModelBuilder(object):
    """
    VersionedModelBuilder handles the building of Version models based on
    parent table attributes and versioning configuration.
    """
    def __init__(self, versioning_manager, model):
        """
        :param versioning_manager:
            VersioningManager object
        :param model:
            SQLAlchemy declarative model object that acts as a parent for the
            built version model
        """
        self.manager = versioning_manager
        self.model = model

    def build_parent_relationship(self):
        """
        Builds a relationship between currently built version class and
        parent class (the model whose history the currently build version
        class represents).
        """
        conditions = []
        foreign_keys = []
        model_keys = []
        for key, column in sa.inspect(self.model).columns.items():
            if column.primary_key:
                conditions.append(
                    getattr(self.model, key)
                    ==
                    getattr(self.version_class, key)
                )
                foreign_keys.append(
                    getattr(self.version_class, key)
                )
                model_keys.append(
                    getattr(self.model, key)
                )

        # We need to check if versions relation was already set for parent
        # class.
        if not hasattr(self.model, 'versions'):
            self.model.versions = sa.orm.relationship(
                self.version_class,
                primaryjoin=sa.and_(*conditions),
                foreign_keys=foreign_keys,
                order_by=lambda: getattr(
                    self.version_class,
                    option(self.model, 'transaction_column_name')
                ),
                lazy='dynamic',
                viewonly=True
            )
            # We must explicitly declare this relationship, instead of
            # specifying as a backref to the one above, since they are
            # viewonly=True and SQLAlchemy will warn if using backref.
            self.version_class.version_parent = sa.orm.relationship(
                self.model,
                primaryjoin=sa.and_(*conditions),
                foreign_keys=model_keys,
                viewonly=True,
                uselist=False,
            )

    def build_transaction_relationship(self, tx_class):
        """
        Builds a relationship between currently built version class and
        Transaction class.

        :param tx_class: Transaction class
        """
        # Only define transaction relation if it doesn't already exist in
        # parent class.

        transaction_column = getattr(
            self.version_class,
            option(self.model, 'transaction_column_name')
        )

        if not hasattr(self.version_class, 'transaction'):
            self.version_class.transaction = sa.orm.relationship(
                tx_class,
                primaryjoin=tx_class.id == transaction_column,
                foreign_keys=[transaction_column],
            )

    def base_classes(self):
        """
        Returns all base classes for history model.
        """
        return (version_base(self.manager, self.model), )

    def inheritance_args(self, cls, version_table, table):
        """
        Return mapper inheritance args for currently built history model.
        """
        args = {}

        if not sa.inspect(self.model).single:
            parent = find_closest_versioned_parent(
                self.manager, self.model
            )
            if parent:
                # The version classes do not contain foreign keys, hence we
                # need to map inheritance condition manually for classes that
                # use joined table inheritance
                if parent.__table__.name != table.name:
                    mapper = sa.inspect(self.model)

                    inherit_condition = adapt_columns(
                        mapper.inherit_condition
                    )
                    tx_column_name = self.manager.options[
                        'transaction_column_name'
                    ]
                    args['inherit_condition'] = sa.and_(
                        inherit_condition,
                        getattr(parent.__table__.c, tx_column_name) ==
                        getattr(cls.__table__.c, tx_column_name)
                    )
                    args['inherit_foreign_keys'] = [
                        version_table.c[column.key]
                        for column in sa.inspect(self.model).columns
                        if column.primary_key
                    ]

        args.update(copy_mapper_args(self.model))

        return args

    def get_inherited_denormalized_columns(self, table):
        parent_models = list(versioned_parents(self.manager, self.model))
        mapper = sa.inspect(self.model)
        args = {}

        if parent_models and not (mapper.single or mapper.concrete):
            columns = [
                self.manager.option(self.model, 'operation_type_column_name'),
                self.manager.option(self.model, 'transaction_column_name')
            ]
            if self.manager.option(self.model, 'strategy') == 'validity':
                columns.append(
                    self.manager.option(
                        self.model,
                        'end_transaction_column_name'
                    )
                )

            for column in columns:
                args[column] = column_property(
                    table.c[column],
                    *[m.__table__.c[column] for m in parent_models]
                )
        return args

    def build_model(self, table):
        """
        Build history model class.
        """
        args = {}

        @declared_attr
        def mapper_args(cls):
            mapper_args = {}
            mapper_args.update(self.inheritance_args(
                cls, table, self.model.__table__)
            )
            return mapper_args

        args['__mapper_args__'] = mapper_args
        args['__versioning_manager__'] = self.manager
        args['__version_parent__'] = self.model

        parent = find_closest_versioned_parent(self.manager, self.model)

        if not parent or parent.__table__.name != table.name:
            args['__table__'] = table

        args.update(self.get_inherited_denormalized_columns(table))

        if self.manager.options.get('use_module_name', True):
            name = '%s%sVersion' % (
                self.model.__module__.title().replace('.', ''),
                self.model.__name__
            )
        else:
            name = '%sVersion' % (self.model.__name__,)
        return type(name, self.base_classes(), args)


    def __call__(self, table, tx_class):
        """
        Build history model and relationships to parent model, transaction
        log model.
        """
        # versioned attributes need to be copied for each child class,
        # otherwise each child class would share the same __versioned__
        # option dict
        self.model.__versioned__ = copy(self.model.__versioned__)
        self.model.__versioning_manager__ = self.manager
        self.version_class = self.build_model(table)
        self.build_parent_relationship()
        self.build_transaction_relationship(tx_class)
        return self.version_class
